import logging

import email_validator
import pydantic

_logger = logging.getLogger(__name__)


class ReallyValidateEmailException(Exception):
    def __init__(self, email: str, error_message: str):
        self.email = email
        self.error_message = error_message

    def __str__(self) -> str:
        return f"{self.__class__.__name__}, {self.email=}, {self.error_message}"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}, {self.email=}, {self.error_message}"


def really_validate_email(email: str, *, log_exception: bool = True):
    try:
        pydantic.validate_email(value=email)
        email_validator.validate_email(email)
    except Exception as exception:
        if log_exception:
            _logger.warning(exception)
        raise ReallyValidateEmailException(
            email=email,
            error_message=str(exception)
        )


def is_really_email_valid(email: str) -> bool:
    try:
        really_validate_email(email=email, log_exception=False)
    except ReallyValidateEmailException:
        return False
    return True


def __example():
    pass


if __name__ == '__main__':
    __example()



# arpakit
from typing import Union, Any

from pydantic import ConfigDict, model_validator
from pydantic_core import PydanticUndefined
from pydantic_settings import BaseSettings


def generate_env_example(settings_class: Union[BaseSettings, type[BaseSettings]]) -> str:
    res = ""
    for k, f in settings_class.model_fields.items():
        if f.default is not PydanticUndefined:
            v = f.default
            # Приводим к строке для .env
            if isinstance(v, bool):
                s = "true" if v else "false"
            elif v is None:
                s = "None"
            else:
                s = str(v)
            # Если дефолт — строка с пробелами → в кавычки (экранируем \ и ")
            if isinstance(v, str) and any(ch.isspace() for ch in v):
                s = '"' + s.replace("\\", "\\\\").replace('"', '\\"') + '"'
            res += f"{k}={s}\n"
        else:
            # обязательное поле — без значения
            res += f"{k}=\n"
    return res


class SimpleSettings(BaseSettings):
    model_config = ConfigDict(extra="ignore")

    @model_validator(mode="before")
    @classmethod
    def validate_all_fields(cls, values: dict[str, Any]) -> dict[str, Any]:
        for key, value in values.items():
            if not isinstance(value, str):
                continue
            if value.lower().strip() in {"null", "none", "nil"}:
                values[key] = None
            elif value.lower().strip() == "default_value":
                field = cls.model_fields.get(key)
                if field is not None:
                    if field.default is not None:
                        values[key] = field.default
                    elif field.default_factory is not None:
                        values[key] = field.default_factory()
        return values

    @classmethod
    def generate_env_example(cls) -> str:
        return generate_env_example(settings_class=cls)

    @classmethod
    def save_env_example_to_file(cls, filepath: str) -> str:
        env_example = cls.generate_env_example()
        with open(filepath, mode="w") as f:
            f.write(env_example)
        return env_example


def __example():
    pass


if __name__ == '__main__':
    __example()



from arpakitlib.ar_json_util import transfer_data_to_json_str


class DictAsObject:
    def __init__(self, data: dict | list):
        self._real_data = data

    def __getattr__(self, item):
        if isinstance(self._real_data, dict):
            try:
                return self.wrap(self._real_data[item])
            except KeyError:
                raise AttributeError(item)

        raise AttributeError(item)

    def __getitem__(self, key):
        if isinstance(self._real_data, (list, dict)):
            return self.wrap(self._real_data[key])
        raise TypeError(f"{type(self._real_data)} is not subscriptable")

    def __len__(self):
        return len(self._real_data)

    def __repr__(self):
        return transfer_data_to_json_str(self._real_data, beautify=True)

    @staticmethod
    def wrap(value):
        if isinstance(value, (dict, list)):
            return DictAsObject(value)
        return value

    def get_raw_from_dict(self, *, key: str, allow_non_exist: bool = True):
        """
        Возвращает значение напрямую из _real_data, без wrap
        """

        if not isinstance(self._real_data, dict):
            raise TypeError("not isinstance(self._real_data, dict)")

        if key not in self._real_data:
            if allow_non_exist is True:
                return None

        return self._real_data[key]

    def __contains__(self, item):
        return item in self._real_data



# arpakit


import asyncio
from datetime import timedelta
from typing import Any

from arpakitlib.ar_sleep_util import sync_safe_sleep, async_safe_sleep


async def async_retry_func(
        *,
        async_func: Any,
        async_func_args: tuple[Any, ...] | None = None,
        async_func_kwargs: dict[str, Any] | None = None,
        max_tries: int = 5,
        endlessly: bool = False,
        timeout_after_exception: timedelta | None = None,
        raise_if_exception: bool = True
):
    tries = 0
    async_func_args = async_func_args or ()
    async_func_kwargs = async_func_kwargs or {}

    while True:
        try:
            return await async_func(*async_func_args, **async_func_kwargs)
        except Exception as exception:
            tries += 1
            if not endlessly and tries >= max_tries:
                if raise_if_exception:
                    raise exception
                return None
            if timeout_after_exception:
                await async_safe_sleep(timeout_after_exception.total_seconds())


def sync_retry_func(
        *,
        sync_func: Any,
        sync_func_args: tuple[Any, ...] | None = None,
        sync_func_kwargs: dict[str, Any] | None = None,
        max_tries: int = 5,
        endlessly: bool = False,
        timeout_after_exception: timedelta | None = None,
        raise_if_exception: bool = True
):
    tries = 0
    sync_func_args = sync_func_args or ()
    sync_func_kwargs = sync_func_kwargs or {}

    while True:
        try:
            return sync_func(*sync_func_args, **sync_func_kwargs)
        except Exception as exception:
            tries += 1
            if not endlessly and tries >= max_tries:
                if raise_if_exception:
                    raise exception
                return None
            if timeout_after_exception:
                sync_safe_sleep(timeout_after_exception.total_seconds())


async def __async_example():
    pass


if __name__ == '__main__':
    asyncio.run(__async_example())



from typing import Any

from pydantic import BaseModel, ConfigDict

from arpakitlib.ar_type_util import NotSet, is_set, is_not_set


class RatFuncRes(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    func_res: Any = NotSet
    func_exception: Exception | None = None

    def raise_for_func_exception(self):
        if self.func_exception is not None:
            raise self.func_exception

    @property
    def has_exception(self) -> bool:
        if self.func_exception is not None:
            return True
        return False

    @property
    def is_func_res_set(self) -> bool:
        return is_set(self.func_res)

    @property
    def is_func_res_not_set(self) -> bool:
        return is_not_set(self.func_res)


def rat_sync_func(
        *,
        sync_func: Any,
        sync_func_args: tuple[Any, ...] | None = None,
        sync_func_kwargs: dict[str, Any] | None = None,
) -> RatFuncRes:
    sync_func_args = sync_func_args or ()
    sync_func_kwargs = sync_func_kwargs or {}
    rat_func_res = RatFuncRes()
    try:
        rat_func_res.func_res = sync_func(*sync_func_args, **sync_func_kwargs)
    except Exception as exception:
        rat_func_res.func_exception = exception
    return rat_func_res


async def rat_async_func(
        *,
        async_func: Any,
        async_func_args: tuple[Any, ...] | None = None,
        async_func_kwargs: dict[str, Any] | None = None,
) -> RatFuncRes:
    async_func_args = async_func_args or ()
    async_func_kwargs = async_func_kwargs or {}
    rat_func_res = RatFuncRes()
    try:
        rat_func_res.func_res = await async_func(*async_func_args, **async_func_kwargs)
    except Exception as exception:
        rat_func_res.func_exception = exception
    return rat_func_res


def __example():
    pass


if __name__ == '__main__':
    __example()



# arpakit
import ast
import asyncio
import json
import logging
from typing import Any, get_origin, get_args, Awaitable, Callable

from aiogram import types
from aiogram.enums import ParseMode
from aiogram.filters import CommandObject
from pydantic import BaseModel, Field, ConfigDict
from pydantic_core import PydanticUndefined

from arpakitlib.ar_parse_command_util import parse_command

_logger = logging.getLogger(__name__)


class BaseTgCommandModel(BaseModel):
    model_config = ConfigDict(extra="ignore", arbitrary_types_allowed=True, from_attributes=True)

    help: bool | None = Field(default=False)


class ExampleTgCommandModel(BaseTgCommandModel):
    """Это описание команды hello-world"""

    hello: str | None = Field(description="Example Desc", default="Hello world")


def _generate_help_text(
        *,
        command_name: str,
        model_class: type[BaseTgCommandModel],
        desc: str | None = None
) -> str:
    """Генерирует help-текст по описанию модели."""
    lines = [f"<b>Command</b> /{command_name}"]

    if desc:
        desc = desc.strip()
        if desc:
            lines.append(f"\n{desc}")

    lines.append(f"\n\n<b>Fields ({len(model_class.model_fields)}):</b>")

    for name, field in model_class.model_fields.items():

        default_value = (
            field.default
            if field.default is not PydanticUndefined
            else "<i>required</i>"
        )

        origin = get_origin(field.annotation)
        args = get_args(field.annotation)

        if origin is list and args:
            type_name = f"list[{args[0].__name__}]"
        elif hasattr(field.annotation, "__name__"):
            type_name = field.annotation.__name__
        else:
            type_name = str(field.annotation)

        description = f" — {field.description}" if field.description else ""

        lines.append(f"\n• <code>-{name}</code>: <i>{type_name}</i>{description} (default: {default_value})")

    lines.append(f"\n\n<b>Usage for help:</b>\n<code>/{command_name} -help</code>")

    return "\n".join(lines)


def _parse_tg_param(value: Any) -> Any:
    """Интеллектуальное преобразование строки из Telegram-команды в Python-значение."""
    if isinstance(value, (bool, int, float)) or value is None:
        return value

    v = value.strip()
    if v.lower() in {"true", "false"}:
        return v.lower() == "true"

    try:
        result = ast.literal_eval(v)
        # Если получился список — проверим, не смесь ли типов
        if isinstance(result, list):
            # Если все элементы одного типа — оставляем
            if len(result) > 0 and len({type(x) for x in result}) > 1:
                # Приводим всё к строкам для единообразия
                result = [str(x) for x in result]
        return result
    except Exception:
        # не получилось распарсить — возвращаем как строку
        pass

    try:
        result = json.loads(v)
        return result
    except Exception:
        pass

    return v


def as_tg_command_handler(
        *,
        tg_command_format_class: type[BaseTgCommandModel],
        desc: str | None = None
) -> Callable[[Callable[..., Awaitable[Any]]], Callable[..., Awaitable[Any]]]:
    def decorator(handler):

        async def new_handler(*args, **kwargs):

            message: types.Message = args[0]
            if not isinstance(message, types.Message):
                raise TypeError("not isinstance(message, types.Message)")

            tg_command: CommandObject = kwargs.get("command")
            if tg_command is None:
                raise ValueError("Missing 'command' in handler kwargs")

            try:

                parsed_command = parse_command(text=tg_command.text)
                kwargs["parsed_command"] = parsed_command

                if parsed_command.has_flag(flag="help") or parsed_command.has_flag(flag="h"):
                    await message.answer(
                        text=_generate_help_text(
                            command_name=tg_command.command,
                            model_class=tg_command_format_class,
                            desc=desc
                        ),
                        disable_web_page_preview=True,
                        parse_mode=ParseMode.HTML
                    )
                    return None

                tg_command_model_data: dict[str, Any] = {}

                for key, value in parsed_command.key_to_value.items():
                    if value is None:
                        tg_command_model_data[key] = True
                    else:
                        tg_command_model_data[key] = _parse_tg_param(value=value)

                for i, value in enumerate(parsed_command.values_without_key):
                    tg_command_model_data[f"arg_{i}"] = _parse_tg_param(value=value)

                tg_command_format_obj = tg_command_format_class(**tg_command_model_data)

                kwargs["tg_command_format_obj"] = tg_command_format_obj

            except Exception as exception:

                _logger.exception("Error while parsing command %s", tg_command.text, exc_info=True)

                await message.answer(
                    text=(
                        f"<b>Bad command usage</b> /{tg_command.command}"
                        f"\n"
                        f"\n<b>Error:</b>"
                        f"\n{exception}"
                        f"\n"
                        f"\nUse <code>/{tg_command.command} -help</code> for help info"
                    ),
                    disable_web_page_preview=True,
                    parse_mode=ParseMode.HTML
                )

                return None

            return await handler(*args, **kwargs)

        return new_handler

    return decorator


def __example():
    pass


async def __async_example():
    pass


if __name__ == '__main__':
    __example()
    asyncio.run(__async_example())



# arpakit

import logging
import os


def init_log_file(*, log_filepath: str | None):
    if not log_filepath:
        return
    directory = os.path.dirname(log_filepath)
    if directory and not os.path.exists(directory):
        os.makedirs(directory, exist_ok=True)
    if not os.path.exists(log_filepath):
        with open(log_filepath, mode="w") as file:
            file.write(" \n")


_normal_easy_logging_was_setup: bool = False


def setup_normal_easy_logging():
    global _normal_easy_logging_was_setup
    if _normal_easy_logging_was_setup:
        logging.getLogger().info("normal easy logging was already setup")
        return

    logger = logging.getLogger()
    logger.setLevel(logging.INFO)

    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(logging.INFO)
    stream_formatter = logging.Formatter(
        "%(asctime)s %(msecs)03d | %(levelname)s | %(name)s | %(filename)s | "
        "%(funcName)s:%(lineno)d - %(message)s",
        datefmt="%d.%m.%Y %H:%M:%S %p %Z %z",
    )
    stream_handler.setFormatter(stream_formatter)
    logger.addHandler(stream_handler)

    _normal_logging_was_setup = True

    logger.info("normal easy logging was setup")


def __example():
    pass


if __name__ == '__main__':
    __example()



# arpakit

from cryptography.fernet import Fernet


def generate_secret_key() -> str:
    return Fernet.generate_key().decode()


def encrypt_with_secret_key(string: str, secret_key: str) -> str:
    return Fernet(secret_key.encode()).encrypt(string.encode()).decode()


def decrypt_with_secret_key(string: str, secret_key: str) -> str:
    return Fernet(secret_key.encode()).decrypt(string.encode()).decode()


def __example():
    pass


if __name__ == '__main__':
    __example()



from urllib.parse import urlparse


class ReallyValidateUrlException(Exception):
    def __init__(self, url: str, error_message: str):
        self.url = url
        self.error_message = error_message

    def __str__(self) -> str:
        return f"{self.__class__.__name__}, {self.url=}, {self.error_message}"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}, {self.url=}, {self.error_message}"


def really_validate_url(url: str) -> None:
    result = urlparse(url)

    if not result.scheme:
        raise ReallyValidateUrlException(
            url=url,
            error_message="URL must include a scheme (e.g., http/https)."
        )

    if not result.netloc:
        raise ReallyValidateUrlException(
            url=url,
            error_message="URL must include a domain or host."
        )

    # Optional: restrict allowed schemes
    if result.scheme not in ("http", "https", "ftp"):
        raise ReallyValidateUrlException(
            url=url,
            error_message=f"Unsupported scheme: {result.scheme}"
        )

    # If everything is fine — return None (no errors)
    return None


def is_really_url_valid(email: str) -> bool:
    try:
        really_validate_url(url=email)
    except ReallyValidateUrlException:
        return False
    return True



# arpakit
from typing import Optional, Any


class NotSet:
    pass


def is_set(v: Any) -> bool:
    return not (v is NotSet or isinstance(v, NotSet))


def is_set_and_not_none(v: Any) -> bool:
    return is_set(v) and (v is not None)


def is_not_set(v: Any) -> bool:
    return not is_set(v)


def is_not_set_or_none(v: Any) -> bool:
    return is_not_set(v) or (v is None)


def raise_if_set(v: Any):
    if is_set(v):
        raise ValueError("value is set")


def raise_if_not_set(v: Any):
    if not is_set(v):
        raise ValueError("value not set")


def make_none_if_not_set(v: Any) -> Any:
    if is_not_set(v):
        return None
    return v


def raise_for_type(comparable, need_type, comment_for_error: Optional[str] = None):
    if comparable is need_type:
        return

    if not isinstance(comparable, need_type):
        err = f"raise_for_type, {comparable} != {need_type}, type {type(comparable)} != type {need_type}"
        if comment_for_error:
            err += f"\n{comment_for_error}"
        raise TypeError(err)


def raise_for_types(comparable, need_types, comment_for_error: Optional[str] = None):
    exceptions = []
    for need_type in need_types:
        try:
            raise_for_type(comparable=comparable, need_type=need_type, comment_for_error=comment_for_error)
            return
        except TypeError as e:
            exceptions.append(e)
    if exceptions:
        err = f"raise_for_types, {comparable} not in {need_types}, type {type(comparable)} not in {need_types}"
        if comment_for_error:
            err += f"\n{comment_for_error}"
        raise TypeError(err)


def make_none_to_false(v: Any) -> Any:
    if v is None:
        return False
    return v


def raise_if_none(v: Any) -> Any:
    if v is None:
        raise ValueError(f"v is None, v={v}, type(v)={type(v)}")
    return v


def raise_if_not_none(v: Any) -> Any:
    if v is not None:
        raise ValueError(f"v is not None, v={v}, type(v)={type(v)}")
    return v


def get_setted_elements_as_dict_from_dict(d: dict) -> dict[str, Any]:
    raise_for_type(d, dict)
    setted_ = {}
    for k, v in d.items():
        if is_set(v):
            setted_[k] = v
    return setted_


def get_setted_keys_from_dict(d: dict) -> list[Any]:
    return list(get_setted_elements_as_dict_from_dict(d).keys())


def __example():
    pass


if __name__ == '__main__':
    __example()



from __future__ import annotations

from typing import Set

import sqlalchemy
from sqlalchemy import CheckConstraint
from sqlalchemy.engine import Engine
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.schema import AddConstraint


# --- Вспомогательные функции ---

def _normalize_sql(sql: str) -> str:
    """
    Нормализует текст SQL выражения для сравнения:
    - нижний регистр
    - убирает лишние пробелы и внешние скобки
    - убирает пробелы вокруг операторов
    """
    s = sql.strip().lower()
    # убираем внешние скобки, если они охватывают всё выражение
    while s.startswith("(") and s.endswith(")"):
        # проверим баланс скобок, чтобы не снимать лишнего
        depth = 0
        balanced = True
        for i, ch in enumerate(s):
            if ch == "(":
                depth += 1
            elif ch == ")":
                depth -= 1
                if depth == 0 and i != len(s) - 1:
                    balanced = False
                    break
        if balanced:
            s = s[1:-1].strip()
        else:
            break
    # схлопываем все пробелы
    import re
    s = re.sub(r"\s+", " ", s)
    # уберем пробелы вокруг простых операторов
    s = re.sub(r"\s*([=<>+\-*/%,])\s*", r"\1", s)
    return s


def _render_check_sql(cc: CheckConstraint, engine: Engine) -> str:
    """Компилируем SQL выражение CHECK в текст с подстановкой литералов."""
    compiled = cc.sqltext.compile(
        dialect=engine.dialect,
        compile_kwargs={"literal_binds": True}
    )
    return str(compiled)


# --- Основная функция ---

def ensure_sqlalchemy_check_constraints(*, base_: type[DeclarativeBase], engine: Engine) -> None:
    """
    Синхронизирует CHECK-ограничения, объявленные в моделях Base.metadata,
    с фактическими ограничениями в базе данных.

    Для каждого CheckConstraint в моделях:
      - если уже есть в БД (по имени или по эквивалентному SQL) — пропускаем
      - иначе создаём (ALTER TABLE ... ADD CONSTRAINT ...)

    Параметры:
      Base  : ваш DeclarativeBase (или его подкласс), от которого наследуются все модели
      engine_or_conn: Engine или Connection SQLAlchemy

    Исключения пробрасываются наверх (например, если СУБД не поддерживает добавление CHECK «на лету»).
    """
    # Возьмём соединение и транзакцию (BEGIN), чтобы изменения были атомарными

    with engine.begin() as conn:
        conn_inspection_ = sqlalchemy.inspect(conn)

        for table in base_.metadata.tables.values():
            if not conn_inspection_.has_table(table.name, schema=table.schema):
                continue

            # Соберём существующие CHECK-и из БД
            existing = conn_inspection_.get_check_constraints(table.name, schema=table.schema)

            # Множества для быстрых проверок
            existing_names: Set[str] = set()
            existing_sql_norms: Set[str] = set()

            for row in existing:
                name = (row.get("name") or "").lower()
                if name:
                    existing_names.add(name)
                sqltext = row.get("sqltext") or row.get("definition") or ""
                if sqltext:
                    existing_sql_norms.add(_normalize_sql(sqltext))

            # Теперь пройдём по CheckConstraint, объявленным в модели
            for constraint in table.constraints:
                if not isinstance(constraint, CheckConstraint):
                    continue

                name = (constraint.name or "").lower()
                # если имя есть и уже существует — готово
                if name and name in existing_names:
                    continue

                # иначе сравним по нормализованному SQL
                try:
                    model_sql = _render_check_sql(constraint, conn)
                    model_sql_norm = _normalize_sql(model_sql)
                except Exception:
                    # если не получилось скомпилировать, попробуем по сырому .sqltext
                    model_sql_norm = _normalize_sql(str(constraint.sqltext))

                if model_sql_norm in existing_sql_norms:
                    continue  # эквивалент уже есть

                # Если сюда дошли — надо создать ограничение
                # Важно: constraint уже «привязан» к таблице (из модели),
                # CreateConstraint сгенерирует корректный ALTER TABLE ... ADD CONSTRAINT ...
                ddl = AddConstraint(constraint)
                conn.execute(ddl)



from urllib.parse import quote_plus


def generate_connection_url(
        *,
        scheme: str = "postgresql",  # общий случай: postgresql, sqlite, redis, amqp и т.п.
        user: str | None = None,
        password: str | None = None,
        host: str | None = "127.0.0.1",
        port: int | None = None,
        database: str | int | None = None,
        quote_query_params: bool = True,
        **query_params
) -> str:
    """
    Универсальная функция для генерации URL соединений (Postgres, Redis, AMQP, SQLite и др.)

    Примеры:
      postgresql://user:pass@localhost:5432/dbname
      sqlite:///path/to/db.sqlite3
      redis://:mypassword@redis:6379/0
      amqp://user:pass@rabbit:5672/myvhost
    """
    # Формируем часть авторизации
    auth_part = ""
    if user and password:
        auth_part = f"{quote_plus(user)}:{quote_plus(password)}@"
    elif password and not user:
        auth_part = f":{quote_plus(password)}@"
    elif user:
        auth_part = f"{quote_plus(user)}@"

    # Формируем хост и порт
    if scheme.startswith("sqlite"):
        host_part = ""
    else:
        host_part = host or ""
        if port:
            host_part += f":{port}"

    # Формируем "базу" (database / номер / путь)
    db_part = ""
    if database is not None:
        db_part = f"/{quote_plus(str(database))}"

    # Формируем query параметры
    query_part = ""
    if query_params:
        query_items = []
        for k, v in query_params.items():
            value = str(v)
            if quote_query_params:
                value = quote_plus(value)
            query_items.append(f"{k}={value}")
        query_part = f"?{'&'.join(query_items)}"

    return f"{scheme}://{auth_part}{host_part}{db_part}{query_part}"


def __example():
    pass


if __name__ == '__main__':
    __example()



# arpakit

from typing import Optional, Any

from arpakitlib.ar_type_util import raise_for_type


def str_in(string: str, main_string: str, *, max_diff: Optional[int] = None) -> bool:
    if string not in main_string:
        return False

    if max_diff is None:
        return True

    diff = len(main_string) - len(string)
    if diff <= max_diff:
        return True

    return False


def bidirectional_str_in(string1: str, string2: str, *, max_diff: Optional[int] = None) -> bool:
    if (
            str_in(string=string1, main_string=string2, max_diff=max_diff)
            or str_in(string=string2, main_string=string1, max_diff=max_diff)
    ):
        return True
    return False


def str_startswith(string: str, main_string: str, max_diff: Optional[int] = None) -> bool:
    if not main_string.startswith(string):
        return False

    if max_diff is None:
        return True

    diff = len(main_string) - len(string)
    if diff <= max_diff:
        return True

    return False


def bidirectional_str_startswith(string1: str, string2: str, max_diff: Optional[int] = None) -> bool:
    if str_startswith(string1, string2, max_diff=max_diff) or str_startswith(string2, string1, max_diff=max_diff):
        return True
    return False


def make_blank_if_none(string: Optional[str] = None) -> str:
    if string is None:
        return ""
    return string


def make_none_if_blank(string: Optional[str] = None) -> str | None:
    if not string:
        return None
    return string


def strip_or_make_none_if_blank(string: Optional[str] = None) -> str | None:
    if string is None:
        return None
    string = string.strip()
    if not string:
        return None
    return string


def none_if_blank(string: Optional[str] = None) -> str | None:
    return make_none_if_blank(string=string)


def lower_and_strip_if_not_none(string: str | None) -> str | None:
    if string is None:
        return None
    return string.lower().strip()


def remove_html(string: str) -> str:
    from bs4 import BeautifulSoup
    raise_for_type(string, str)
    return BeautifulSoup(string, "html.parser").text


def remove_markdown(text: str) -> str:
    import markdown
    from bs4 import BeautifulSoup
    html_text = markdown.markdown(text)
    soup = BeautifulSoup(html_text, "html.parser")
    return soup.get_text(separator=" ", strip=True)


def remove_tags(string: str) -> str:
    raise_for_type(string, str)
    return string.replace("<", "").replace(">", "")


def remove_tags_and_html(string: str) -> str:
    raise_for_type(string, str)
    return remove_tags(remove_html(string))


def raise_if_string_blank(string: str) -> str:
    if not string:
        raise ValueError("string is blank")
    return string


def return_str_if_none(value: str | None, value_str: str) -> str:
    if value is None:
        return value_str
    return value


def strip_if_not_none(v: str | None) -> str | None:
    if v is None:
        return v
    return v.strip()


def strip_or_none(value: Any):
    if isinstance(value, str):
        value = value.strip()
        if not value:
            return None
        return value
    return None


def lower_and_strip_or_none(value: Any) -> str | None:
    value = strip_or_none(value)
    if isinstance(value, str):
        return value.lower().strip()
    return None


def is_float(value: str) -> bool:
    try:
        float(value)
    except Exception:
        return False
    return True


def int_or_float_or_none(value: Any) -> float | int | None:
    if isinstance(value, float) or isinstance(value, int):
        return value
    value = strip_or_none(value)
    if value is not None and value.isdigit():
        return int(value)
    if value is not None and is_float(value):
        return float(value)
    return None


def int_or_none(value: Any) -> int | None:
    if isinstance(value, int):
        return value
    value = strip_or_none(value)
    if value is not None and value.isdigit():
        return int(value)
    return None


def float_or_none(value: Any) -> float | None:
    if isinstance(value, float):
        return value
    value = strip_or_none(value)
    if value is not None and is_float(value):
        return float(value)
    return None


def __example():
    pass


if __name__ == '__main__':
    __example()



# arpakit
import datetime as dt
from typing import Any

from pydantic import BaseModel, ConfigDict, Field

from arpakitlib.ar_datetime_util import now_utc_dt


class SafeFuncResult(BaseModel):
    model_config = ConfigDict(extra="forbid", arbitrary_types_allowed=True, from_attributes=True)

    has_exception: bool = False
    func_result: Any = None
    exception: Exception | None = None
    duration: dt.timedelta | None = None
    func_name: str
    args: tuple = Field(default_factory=tuple)
    kwargs: dict = Field(default_factory=dict)

    @property
    def is_ok(self) -> bool:
        if self.has_exception:
            return False
        return True

    def simple_dict_for_json(self) -> dict[str, Any]:
        return {
            "has_exception": self.has_exception,
            "func_result": self.func_result,
            "exception": self.exception,
            "duration": self.duration,
            "duration_total_seconds": self.duration.total_seconds() if self.duration is not None else None,
            "func_name": self.func_name,
            "is_ok": self.is_ok
        }


def sync_safely_run_func(*, sync_func, args: tuple | None = None, kwargs: dict | None = None) -> SafeFuncResult:
    if args is None:
        args = tuple()
    if kwargs is None:
        kwargs = {}
    func_start_dt = now_utc_dt()
    try:
        res = sync_func(*args, **kwargs)
        duration = now_utc_dt() - func_start_dt
        return SafeFuncResult(
            has_exception=False,
            func_result=res,
            duration=duration,
            func_name=sync_func.__name__,
            args=args,
            kwargs=kwargs
        )
    except Exception as exception:
        duration = now_utc_dt() - func_start_dt
        return SafeFuncResult(
            has_exception=True,
            exception=exception,
            duration=duration,
            func_name=sync_func.__name__,
            args=args,
            kwargs=kwargs
        )


def __example():
    pass


if __name__ == "__main__":
    __example()



import importlib.util
import os

from fastapi import APIRouter


def include_fastapi_routers_from_dir(
        *,
        router: APIRouter,
        base_dir: str = ".",
        exclude_filenames: list[str] | None = None,
):
    """
    Рекурсивно ищет все .py файлы с объектом `api_router` типа APIRouter
    и подключает их к переданному `router`.

    Префикс = имя файла без .py
    exclude_filenames — список имён файлов, которые нужно пропустить (без путей)
    """
    if exclude_filenames is None:
        exclude_filenames = ["__init__.py"]

    for root, _, files in os.walk(base_dir):
        files.sort()

        for filename in files:
            if not filename.endswith(".py") or filename in exclude_filenames:
                continue

            file_path = os.path.join(root, filename)
            module_name = (
                os.path.relpath(file_path, base_dir)
                .replace(os.sep, ".")
                .removesuffix(".py")
            )

            spec = importlib.util.spec_from_file_location(module_name, file_path)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)

            api_router = getattr(module, "api_router", None)
            if isinstance(api_router, APIRouter):
                prefix = "/" + filename[:-3]  # имя файла без .py
                router.include_router(
                    router=api_router,
                    prefix=prefix
                )



# arpakit

import asyncio
import math
from datetime import timedelta
from time import sleep


def sync_safe_sleep(n: timedelta | float | int):
    if isinstance(n, timedelta):
        n = n.total_seconds()
    elif isinstance(n, int):
        n = float(n)
    elif isinstance(n, float):
        n = n
    else:
        raise TypeError(f"n={n}, type={type(n)}, n: timedelta | float | int")

    n: float = n

    frac, int_part = math.modf(n)
    for i in range(int(int_part)):
        sleep(1)
    sleep(frac)


async def async_safe_sleep(n: timedelta | float | int):
    if isinstance(n, timedelta):
        n = n.total_seconds()
    elif isinstance(n, int):
        n = float(n)
    elif isinstance(n, float):
        n = n
    else:
        raise TypeError(f"n={n}, type={type(n)}, n: timedelta | float | int")

    n: float = n

    await asyncio.sleep(n)


def __example():
    pass


async def __async_example():
    pass


if __name__ == '__main__':
    __example()
    asyncio.run(__async_example())



from typing import Any


def _str_no_self(v: str) -> str:
    return v.replace("self.", "").strip()


def _str_no_dbm(v: str) -> str:
    return v.replace("dbm.", "").strip()


def generate_sqlalchemy_model_repr(*, title: str, parts: list[Any]):
    res = []
    for part in parts:
        if part is None:
            continue
        if isinstance(part, int):
            part = str(part)
        elif isinstance(part, bool):
            part = str(part)
        elif isinstance(part, str):
            part = _str_no_self(part)
            part = _str_no_dbm(part)
            part = part.strip()
        if not part:
            continue
        res.append(part)
    return f"{title} ({', '.join(res)})"



# arpakit

from typing import Any, Optional

import jwt
from jwt import PyJWTError

from arpakitlib.ar_type_util import raise_for_type


def encode_jwt_token(
        *,
        jwt_payload: dict[str, Any],
        jwt_secret: str
) -> str:
    raise_for_type(jwt_secret, str)
    return jwt.encode(jwt_payload, jwt_secret, algorithm="HS256")


def decode_jwt_token(
        *,
        jwt_token: str,
        jwt_secret: str
) -> Optional[dict[str, Any]]:
    raise_for_type(jwt_token, str)
    raise_for_type(jwt_secret, str)
    try:
        return jwt.decode(jwt_token, jwt_secret, algorithms=["HS256"])
    except PyJWTError:
        return None


def __example():
    pass


if __name__ == '__main__':
    __example()



# arpakit

from typing import Any


def combine_dicts(*dicts: dict) -> dict[Any, Any]:
    res = {}
    for dict_ in dicts:
        res.update(dict_)
    return res


def replace_dict_key(*, dict_: dict, old_key: Any, new_key: Any) -> dict[Any, Any]:
    if old_key in dict_:
        dict_[new_key] = dict_.pop(old_key)
    return dict_


class GetTypedFromDictException(Exception):
    pass


def get_typed_from_dict(
        *,
        d: dict,
        key: str,
        type_,
        allow_missing: bool = False,
        custom_exception: type[Exception] = GetTypedFromDictException
) -> Any:
    """
    Получает d[key], проверяет, что оно относится к типу typ.
    Если ключ отсутствует:
        - если allow_non_existing=True → возвращаем None
        - иначе → KeyError
    Если тип не совпадает → TypeError
    """

    if key not in d:
        if allow_missing:
            return None
        raise custom_exception(f"Missing key: {key}")

    val = d[key]

    if not isinstance(val, type_):
        raise custom_exception(f"Expected {type_}, got {type(val)} for key '{key}'")

    return val


def sort_dict_by_keys(*, data: dict[str, Any]) -> dict[str, Any]:
    return {k: data[k] for k in sorted(data)}


def __example():
    pass


if __name__ == '__main__':
    __example()



import logging
from typing import Any

import aiohttp
import requests

from arpakitlib.ar_dict_util import combine_dicts
from arpakitlib.ar_http_request_util import async_make_http_request, sync_make_http_request


class BaseHTTPAPIClient:
    def __init__(self):
        self.headers = {"Content-Type": "application/json"}
        self._logger = logging.getLogger(f"{self.__class__.__name__}")

    def _sync_make_http_request(
            self,
            *,
            method: str,
            url: str,
            headers: dict[str, Any] | None = None,
            **kwargs
    ) -> requests.Response:
        return sync_make_http_request(
            method=method,
            url=url,
            headers=combine_dicts(self.headers, (headers if headers is not None else {})),
            **kwargs
        )

    async def _async_make_http_request(
            self,
            *,
            method: str = "GET",
            url: str,
            headers: dict[str, Any] | None = None,
            **kwargs
    ) -> aiohttp.ClientResponse:
        return await async_make_http_request(
            method=method,
            url=url,
            headers=combine_dicts(self.headers, (headers if headers is not None else {})),
            **kwargs
        )

    def healthcheck(self) -> bool:
        raise NotImplemented()

    async def async_healthcheck(self) -> bool:
        raise NotImplemented()



# arpakit

import logging
from datetime import datetime
from typing import Optional, Any

import pytz
from pydantic import BaseModel

from arpakitlib.ar_json_db_util import JSONDbFile
from arpakitlib.ar_type_util import raise_for_type


class CacheBlock(BaseModel):
    key: str
    data: Any
    last_update_dt: datetime


class CacheFile:

    def __init__(self, *, json_db_file: JSONDbFile):
        self.json_db_file = json_db_file
        self._logger = logging.getLogger(self.__class__.__name__)

    def __len__(self) -> int:
        return self.json_db_file.count_records()

    def __str__(self) -> str:
        return f"CacheFile ({self.json_db_file.filepath}) ({self.json_db_file.count_records()})"

    def __repr__(self) -> str:
        return f"CacheFile ({self.json_db_file.filepath}) ({self.json_db_file.count_records()})"

    def create_block(
            self,
            *,
            key: str,
            data: Any,
            last_update_dt: Optional[datetime] = None
    ) -> CacheBlock:
        raise_for_type(key, str)

        if last_update_dt is None:
            last_update_dt = datetime.now(tz=pytz.UTC)
        raise_for_type(last_update_dt, datetime)
        last_update_dt = last_update_dt.astimezone(tz=pytz.UTC)

        _, record = self.json_db_file.create_record(
            record_id=key,
            record={
                "data": data,
                "last_update_dt": last_update_dt.isoformat()
            }
        )

        return CacheBlock(
            key=key,
            data=record["data"],
            last_update_dt=datetime.fromisoformat(record["last_update_dt"])
        )

    def get_block(self, key: str) -> Optional[CacheBlock]:
        raise_for_type(key, str)

        record = self.json_db_file.get_record(record_id=key)
        if record is None:
            return None

        return CacheBlock(
            key=key,
            data=record["data"],
            last_update_dt=datetime.fromisoformat(record["last_update_dt"])
        )

    def get_blocks(self) -> list[CacheBlock]:
        return [
            CacheBlock(
                key=record_id,
                data=record["data"],
                last_update_dt=datetime.fromisoformat(record["last_update_dt"])
            )
            for record_id, record in self.json_db_file.get_records()
        ]

    def update_block(
            self,
            *,
            key: str,
            data: Optional[dict[str, Any]] = None,
            last_update_dt: Optional[datetime] = None
    ) -> CacheBlock:
        raise_for_type(key, str)
        record = self.json_db_file.get_record(record_id=key)
        if record is None:
            raise ValueError(f"block (key='{key}') not exists")

        if data is not None:
            raise_for_type(data, dict)
            record["data"] = data

        if last_update_dt is not None:
            raise_for_type(last_update_dt, datetime)
            last_update_dt = last_update_dt.astimezone(tz=pytz.UTC)
            record["last_update_dt"] = last_update_dt.isoformat()

        self.json_db_file.update_record(record_id=key, record=record)

        return self.get_block(key=key)

    def remove_block(self, key: str):
        raise_for_type(key, str)
        self.json_db_file.rm_record(record_id=key)

    def remove_blocks(self):
        self.json_db_file.rm_all_records()


def __example():
    pass


if __name__ == '__main__':
    __example()



# arpakit

from __future__ import annotations

import asyncssh
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ed25519
from pydantic import BaseModel, Field


class SSHKeys(BaseModel):
    private_key: str = Field()
    public_key: str = Field()


def generate_ed25519_via_asyncssh_ssh_keys() -> SSHKeys:
    key = asyncssh.generate_private_key("ssh-ed25519")

    private_key_str = key.export_private_key().decode()
    public_key_str = key.export_public_key().decode()

    return SSHKeys(
        private_key=private_key_str,
        public_key=public_key_str
    )


def generate_ed25519_via_cryptography_ssh_keys() -> SSHKeys:
    private_key = ed25519.Ed25519PrivateKey.generate()

    private_bytes = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.OpenSSH,
        encryption_algorithm=serialization.NoEncryption()
    )

    public_key = private_key.public_key().public_bytes(
        encoding=serialization.Encoding.OpenSSH,
        format=serialization.PublicFormat.OpenSSH
    )

    return SSHKeys(
        private_key=private_bytes.decode(),
        public_key=public_key.decode()
    )



from __future__ import annotations

import re
from pathlib import Path
from typing import Optional

LINE_RE = re.compile(r"""
    ^(?P<lead>\s*)
    (?:(?P<export>export)\s+)?        # optional 'export '
    (?P<key>[A-Za-z_][A-Za-z0-9_]*?)  # KEY
    (?P<eq>\s*=\s*)                   # equals with any spaces around
    (?P<val>.*)                       # value (raw, keep as-is)
    $
""", re.VERBOSE)

COMMENT_KEY_RE = re.compile(r"^(\s*#\s*)([A-Za-z_][A-Za-z0-9_]*)(\s*=\s*.*)$")


def uppercase_env_keys(
        *, path: str | Path,
        output: Optional[str | Path] = None,
        backup: bool = False
) -> Path:
    """
    Преобразует имена переменных в .env и в комментариях в верхний регистр.

    :param path: путь к исходному .env
    :param output: путь к файлу вывода. Если None — правит на месте.
    :param backup: делать .bak при in-place
    :return: путь к записанному файлу
    """
    src = Path(path)
    if not src.exists():
        raise FileNotFoundError(src)

    text = src.read_text(encoding="utf-8-sig").splitlines(keepends=True)

    out_lines = []

    for line in text:
        stripped = line.strip()

        # пустые строки — оставляем как есть
        if not stripped:
            out_lines.append(line)
            continue

        # комментарий с "ключом" вида # KEY=
        m_comment = COMMENT_KEY_RE.match(line.rstrip("\n"))
        if m_comment:
            prefix, key, rest = m_comment.groups()
            key_up = key.upper()
            # сохраняем оригинальный символ конца строки, если он был
            newline = "\n" if line.endswith("\n") else ""
            out_lines.append(f"{prefix}{key_up}{rest}{newline}")
            continue

        # стандартная строка с KEY=VAL
        m = LINE_RE.match(line.rstrip("\n"))
        if not m:
            out_lines.append(line)
            continue

        lead = m.group("lead") or ""
        export_kw = m.group("export") or ""
        key = m.group("key")
        eq = m.group("eq")
        val = m.group("val")

        key_up = key.upper()
        newline = "\n" if line.endswith("\n") else ""
        new_line = f"{lead}{(export_kw + ' ') if export_kw else ''}{key_up}{eq}{val}{newline}"
        out_lines.append(new_line)

    # запись
    if output is None:
        if backup:
            bak = src.with_suffix(src.suffix + ".bak")
            bak.write_text("".join(text), encoding="utf-8")
        dst = src
    else:
        dst = Path(output)

    dst.write_text("".join(out_lines), encoding="utf-8")
    return dst



# arpakit

from typing import Union, Iterator, Iterable, Any

ValueType = Union[int, str]
ValuesForParseType = Union[ValueType, Iterable[ValueType]]


class Enumeration:
    @classmethod
    def iter_values(cls) -> Iterator[ValueType]:
        big_dict = {}
        for class_ in reversed(cls.mro()):
            big_dict.update(class_.__dict__)
        big_dict.update(cls.__dict__)

        keys = list(big_dict.keys())

        for key in keys:

            if not isinstance(key, str):
                continue

            if key.startswith("__") or key.endswith("__"):
                continue

            value = big_dict[key]
            if type(value) not in [str, int]:
                continue

            yield value

    @classmethod
    def values_set(cls) -> set[ValueType]:
        return set(cls.iter_values())

    @classmethod
    def values_list(cls) -> list[ValueType]:
        return list(cls.iter_values())

    @classmethod
    def parse_values(cls, *values: ValuesForParseType, validate: bool = False) -> list[ValueType]:
        res = []

        for value in values:

            if isinstance(value, str) or isinstance(value, int):
                if validate is True and value not in cls.values_set():
                    raise ValueError(f"{value} not in {cls.values_set()}")
                res.append(value)

            elif isinstance(value, Iterable):
                for value_ in value:
                    if isinstance(value_, str) or isinstance(value_, int):
                        if validate is True and value_ not in cls.values_set():
                            raise ValueError(f"{value_} not in {cls.values_set()}")
                        res.append(value_)
                    else:
                        raise TypeError(f"bad type, value={value}, type={type(value)}")

            else:
                raise TypeError(f"bad type, value={value}, type={type(value)}")

        return res

    @classmethod
    def parse_and_validate_values(cls, *values: ValuesForParseType) -> list[ValueType]:
        return cls.parse_values(*values, validate=True)

    @classmethod
    def str_for_print(cls) -> str:
        res = f"{cls.__name__} (len={len(cls.values_list())})"
        for v in cls.values_list():
            res += f"\n- {v}"
        return res.strip()

    @classmethod
    def print(cls):
        print(cls.str_for_print())

    @classmethod
    def key_to_value(cls) -> dict[str, Any]:
        big_dict = {}
        for class_ in reversed(cls.mro()):
            big_dict.update(class_.__dict__)
        big_dict.update(cls.__dict__)

        result = {}
        for key, value in big_dict.items():
            if (
                    isinstance(key, str)
                    and not (key.startswith("__") or key.endswith("__"))
                    and isinstance(value, (str, int))
            ):
                result[key] = value

        return result


def __example():
    pass


if __name__ == '__main__':
    __example()



# arpakit

from typing import Any


def iter_group_list(list_: list[Any], n: int):
    part = []
    for v in list_:
        if len(part) < n:
            part.append(v)
        else:
            yield part.copy()
            part = [v]
    yield part


def group_list(list_: list[Any], n: int):
    return list(iter_group_list(list_=list_, n=n))


def remove_from_list_if_left(*, list_: list[Any], values: list[Any]) -> list[Any]:
    """
    Удаляет из list_ все элементы, чьи значения перечислены в values,
    НО только если после удаления в list_ останется хотя бы один элемент.

    Всегда возвращает сам список list_ (после возможной модификации in-place).
    """
    if not values:
        return list_

    targets = set(values)
    # кандидат на новое содержимое
    filtered = [x for x in list_ if x not in targets]

    # если нечего удалять — оставляем как есть
    if len(filtered) == len(list_):
        return list_

    # удаляем только если список не опустеет
    if len(filtered) == 0:
        return list_

    # применяем изменения in-place
    list_[:] = filtered
    return list_


def remove_from_lists_if_left(
        *,
        lists_: list[list[Any]],
        values: list[Any]
):
    for list_ in lists_:
        remove_from_list_if_left(list_=list_, values=values)
    return lists_


def int_to_list_(value: int | list | None) -> list[int] | None:
    if isinstance(value, int):
        return [value]
    elif isinstance(value, list):
        return value
    elif value is None:
        return None
    else:
        raise TypeError(f"{type(value)=}")


def str_to_list_(value: str | list | None) -> list[str] | None:
    if isinstance(value, str):
        return [value]
    elif isinstance(value, list):
        return value
    elif value is None:
        return None
    else:
        raise TypeError(f"{type(value)=}")


def __example():
    pass


if __name__ == '__main__':
    __example()



from __future__ import annotations

import asyncio
import logging

import httpx
from openai import OpenAI, AsyncOpenAI
from openai.types.chat import ChatCompletion

from arpakitlib.ar_base64_util import convert_file_to_base64_string

"""
https://platform.openai.com/docs/
"""


class EasyOpenAIAPIClient:
    def __init__(
            self,
            *,
            open_ai: OpenAI,
            async_open_ai: AsyncOpenAI
    ):
        self._logger = logging.getLogger(self.__class__.__name__)

        self.open_ai = open_ai
        self.async_open_ai = async_open_ai

    @classmethod
    def create_easily(
            cls,
            openai_api_key: str,
            openai_api_base_url: str | None = "https://api.proxyapi.ru/openai/v1"
    ) -> EasyOpenAIAPIClient:
        return EasyOpenAIAPIClient(
            open_ai=OpenAI(
                api_key=openai_api_key,
                base_url=openai_api_base_url,
                timeout=httpx.Timeout(
                    timeout=300,  # общий таймаут
                    connect=5,  # таймаут подключения
                    read=300,  # чтение ответа
                    write=60,  # запись запроса
                    pool=10
                )
            ),
            async_open_ai=AsyncOpenAI(
                api_key=openai_api_key,
                base_url=openai_api_base_url,
                http_client=httpx.AsyncClient(
                    timeout=httpx.Timeout(
                        timeout=300,
                        connect=5,
                        read=300,
                        write=60,
                        pool=10
                    )
                )
            )
        )

    def check_conn(self):
        self.open_ai.models.list()

    def is_conn_good(self) -> bool:
        try:
            self.check_conn()
            return True
        except Exception as e:
            self._logger.error(e)
        return False

    async def async_check_conn(self):
        await self.async_open_ai.models.list()

    async def async_is_conn_good(self) -> bool:
        try:
            await self.async_check_conn()
            return True
        except Exception as e:
            self._logger.error(e)
        return False

    def simple_ask(
            self,
            *,
            prompt: str | None = None,
            text: str,
            model: str = "gpt-4o",
            image_links: str | list[str] | None = None,
            image_filepaths: str | list[str] | None = None
    ) -> ChatCompletion:
        if isinstance(image_links, str):
            image_links = [image_links]
        if isinstance(image_filepaths, str):
            image_filepaths = [image_filepaths]

        messages = []

        if prompt is not None:
            messages.append({
                "role": "system",
                "content": prompt
            })

        content = [{"type": "text", "text": text}]

        if image_links:
            for link in image_links:
                content.append({
                    "type": "image_url",
                    "image_url": {"url": link}
                })

        if image_filepaths:
            for path in image_filepaths:
                base64_url = convert_file_to_base64_string(filepath=path, raise_for_error=True)
                content.append({
                    "type": "image_url",
                    "image_url": {"url": base64_url}
                })

        messages.append({
            "role": "user",
            "content": content
        })

        response: ChatCompletion = self.open_ai.chat.completions.create(
            model=model,
            messages=messages,
            n=1,
            temperature=0.1,
            top_p=0.9,
            max_tokens=1000
        )

        return response

    async def async_simple_ask(
            self,
            *,
            prompt: str | None = None,
            text: str,
            model: str = "gpt-4o",
            image_links: str | list[str] | None = None,
            image_filepaths: str | list[str] | None = None
    ) -> ChatCompletion:
        if isinstance(image_links, str):
            image_links = [image_links]
        if isinstance(image_filepaths, str):
            image_filepaths = [image_filepaths]

        messages = []

        if prompt is not None:
            messages.append({
                "role": "system",
                "content": prompt
            })

        content = [{"type": "text", "text": text}]

        if image_links:
            for link in image_links:
                content.append({
                    "type": "image_url",
                    "image_url": {"url": link}
                })

        if image_filepaths:
            for path in image_filepaths:
                base64_url = convert_file_to_base64_string(filepath=path, raise_for_error=True)
                content.append({
                    "type": "image_url",
                    "image_url": {"url": base64_url}
                })

        messages.append({
            "role": "user",
            "content": content
        })

        response: ChatCompletion = await self.async_open_ai.chat.completions.create(
            model=model,
            messages=messages,
            n=1,
            temperature=0.1,
            top_p=0.9,
            max_tokens=1000
        )

        return response


def __example():
    pass


async def __async_example():
    pass


if __name__ == '__main__':
    __example()
    asyncio.run(__async_example())



# arpakit
from typing import Any, Type, Iterable

from pydantic import BaseModel, create_model
from pydantic_core import PydanticUndefined


def clone_pydantic_model_fields(
        *,
        model_cls: Type[BaseModel],
        base_model: Type[BaseModel] = BaseModel,
        fields_to_remove: Iterable[str] | None = None,
        new_class_name: str | None = None,
        class_name_suffix: str | None = "Cloned"
) -> Type[BaseModel]:
    if fields_to_remove is None:
        fields_to_remove = set()
    if class_name_suffix is None:
        class_name_suffix = "Cloned"
    if new_class_name is None:
        new_class_name = f"{model_cls.__name__}{class_name_suffix}"

    field_defs: dict[str, tuple[type[Any], Any]] = {}

    for field_name, field_ in model_cls.model_fields.items():
        if field_name in fields_to_remove:
            continue

        if field_.default_factory is not None and field_.default is PydanticUndefined:
            default = field_
        elif field_.default is not PydanticUndefined:
            default = field_.default
        else:
            default = field_

        field_defs[field_name] = ((field_.annotation or Any), default)

    res = create_model(
        new_class_name,
        __base__=base_model,
        __module__=model_cls.__module__,
        **field_defs,
    )

    return res


def __example():
    pass


if __name__ == '__main__':
    __example()



# arpakit

import asyncio
import logging
from datetime import timedelta
from typing import Any

import aiohttp
import requests
from aiohttp_socks import ProxyConnector

from arpakitlib.ar_json_util import transfer_data_to_json_str
from arpakitlib.ar_sleep_util import sync_safe_sleep, async_safe_sleep
from arpakitlib.ar_type_util import raise_for_type

_logger = logging.getLogger(__name__)


def sync_make_http_request(
        *,
        method: str = "GET",
        url: str,
        headers: dict[str, Any] | None = None,
        params: dict[str, Any] | None = None,
        max_tries_: int = 9,
        proxy_url_: str | None = None,
        raise_for_status_: bool = False,
        not_raise_for_statuses_: list[int] | set[int] | None = None,
        timeout_: timedelta | float = timedelta(seconds=15).total_seconds(),
        enable_logging_: bool = False,
        exception_class_: type[Exception] | None = None,
        retry_delay_timeout: timedelta | int | None = timedelta(seconds=0.1),
        **kwargs
) -> requests.Response:
    if isinstance(retry_delay_timeout, int):
        retry_delay_timeout = timedelta(seconds=retry_delay_timeout)
    if retry_delay_timeout is not None:
        raise_for_type(retry_delay_timeout, timedelta)

    if isinstance(timeout_, float):
        timeout_ = timedelta(seconds=timeout_)
    raise_for_type(timeout_, timedelta)

    tries_counter = 0

    kwargs["method"] = method
    kwargs["url"] = url
    if headers is not None:
        kwargs["headers"] = headers
    if params is not None:
        kwargs["params"] = params
    if proxy_url_:
        kwargs["proxies"] = {
            "http": proxy_url_,
            "https": proxy_url_
        }
    if timeout_ is not None:
        kwargs["timeout"] = timeout_.total_seconds()
    if "allow_redirects" not in kwargs:
        kwargs["allow_redirects"] = True

    if enable_logging_:
        _logger.info(f"try http {method} {url} {params}")

    while True:
        tries_counter += 1
        try:
            response = requests.request(**kwargs)
            if raise_for_status_:
                if not_raise_for_statuses_ and response.status_code in not_raise_for_statuses_:
                    if enable_logging_:
                        _logger.info(f"ignored status {response.status_code} {method} {url} {params}")
                else:
                    try:
                        response.raise_for_status()
                    except requests.HTTPError:
                        if enable_logging_:
                            try:
                                json_data = response.json()
                                _logger.error(
                                    f"bad status {method} {url} {params}"
                                    f"{transfer_data_to_json_str(data=json_data, beautify=True)}"
                                )
                            except Exception:
                                try:
                                    text = response.text
                                    _logger.error(
                                        f"bad status {method} {url} {params}"
                                        f"{text}"
                                    )
                                except Exception:
                                    pass
                        raise
            if enable_logging_:
                _logger.info(f"good try http {method} {url} {params}")
            return response
        except Exception as exception:
            if enable_logging_:
                _logger.warning(
                    f"{tries_counter}/{max_tries_}, bad try {method} {url} {params}, exception={exception}"
                )
            if tries_counter >= max_tries_:
                if exception_class_ is not None:
                    raise exception_class_(exception)
                else:
                    raise
            if retry_delay_timeout is not None:
                sync_safe_sleep(retry_delay_timeout.total_seconds())
            continue


async def async_make_http_request(
        *,
        method: str = "GET",
        url: str,
        headers: dict[str, Any] | None = None,
        params: dict[str, Any] | None = None,
        max_tries_: int = 9,
        proxy_url_: str | None = None,
        raise_for_status_: bool = False,
        not_raise_for_statuses_: list[int] | set[int] | None = None,
        timeout_: timedelta | None = timedelta(seconds=15),
        enable_logging_: bool = False,
        exception_class_: type[Exception] | None = None,
        retry_delay_timeout: timedelta | int | None = timedelta(seconds=0.1),
        **kwargs
) -> aiohttp.ClientResponse:
    if isinstance(retry_delay_timeout, int):
        retry_delay_timeout = timedelta(seconds=retry_delay_timeout)
    if retry_delay_timeout is not None:
        raise_for_type(retry_delay_timeout, timedelta)

    if isinstance(timeout_, float):
        timeout_ = timedelta(seconds=timeout_)
    raise_for_type(timeout_, timedelta)

    tries_counter = 0

    kwargs["method"] = method
    kwargs["url"] = url
    if headers is not None:
        kwargs["headers"] = headers
    if params is not None:
        kwargs["params"] = params
    if timeout_ is not None:
        kwargs["timeout"] = aiohttp.ClientTimeout(total=timeout_.total_seconds())
    if "allow_redirects" not in kwargs:
        kwargs["allow_redirects"] = True

    proxy_connector: ProxyConnector | None = None
    if proxy_url_:
        proxy_connector = ProxyConnector.from_url(proxy_url_)

    if enable_logging_:
        _logger.info(f"try http {method} {url} {params}")

    while True:
        tries_counter += 1
        try:
            async with aiohttp.ClientSession(connector=proxy_connector) as session:
                async with session.request(**kwargs) as response:
                    if raise_for_status_:
                        if not_raise_for_statuses_ and response.status in not_raise_for_statuses_:
                            if enable_logging_:
                                _logger.info(
                                    f"ignored status {response.status} {method} {url} {params}"
                                )
                        else:
                            try:
                                response.raise_for_status()
                            except aiohttp.ClientResponseError:
                                if enable_logging_:
                                    try:
                                        json_data = await response.json()
                                        _logger.error(
                                            f"bad status {method} {url} {params}"
                                            f"{transfer_data_to_json_str(data=json_data, beautify=True)}"
                                        )
                                    except Exception:
                                        try:
                                            text = await response.text()
                                            _logger.error(
                                                f"bad status {method} {url} {params}"
                                                f"{text}"
                                            )
                                        except Exception:
                                            pass
                                raise
                    await response.read()
                    if enable_logging_:
                        _logger.info(f"good try {method} {url} {params}")
                    return response
        except Exception as exception:
            if enable_logging_:
                _logger.warning(
                    f"{tries_counter}/{max_tries_}, bad try {method} {url} {params}, exception={exception}"
                )
            if tries_counter >= max_tries_:
                if exception_class_ is not None:
                    raise exception_class_(exception)
                else:
                    raise
            if retry_delay_timeout is not None:
                await async_safe_sleep(retry_delay_timeout.total_seconds())
            continue


def __example():
    pass


async def __async_example():
    pass


if __name__ == '__main__':
    __example()
    asyncio.run(__async_example())



import asyncio
import inspect
from typing import Callable


def is_async_func(func: Callable) -> bool:
    if inspect.ismethod(func) or inspect.isfunction(func):
        return inspect.iscoroutinefunction(func)
    if isinstance(func, (staticmethod, classmethod)):
        return inspect.iscoroutinefunction(func.__func__)
    return False


def raise_if_not_async_func(func: Callable):
    if not is_async_func(func):
        raise TypeError(f"the provided callable '{func.__name__}' is not an async")


# ---


def is_coroutine(obj: object) -> bool:
    return asyncio.iscoroutine(obj)


# ---


def is_sync_func(func: Callable) -> bool:
    return callable(func) and not is_async_func(func=func)


def raise_if_not_sync_func(func: Callable):
    if not is_sync_func(func):
        raise TypeError(f"the provided callable '{func.__name__}' is not an sync")


#


def get_func_name(func: Callable) -> str:
    return func.__name__


def __example():
    pass


if __name__ == '__main__':
    __example()



import importlib.util
import os

from aiogram import Router


def include_aiogram_routers_from_dir(
        *,
        router: Router,
        base_dir: str = ".",
        exclude_filenames: list[str] | None = None,
):
    if exclude_filenames is None:
        exclude_filenames = ["__init__.py"]

    for root, _, files in os.walk(base_dir):
        files.sort()

        for filename in files:
            if not filename.endswith(".py") or filename in exclude_filenames:
                continue

            file_path = os.path.join(root, filename)
            module_name = (
                os.path.relpath(file_path, base_dir)
                .replace(os.sep, ".")
                .removesuffix(".py")
            )

            spec = importlib.util.spec_from_file_location(module_name, file_path)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)

            tg_bot_router = getattr(module, "tg_bot_router", None)
            if isinstance(tg_bot_router, Router):
                router.include_router(
                    router=tg_bot_router,
                )



# arpakit

import json
import logging
import os
import shutil
from typing import Any, Optional, Self

from arpakitlib.ar_type_util import raise_for_type


class JSONDbFile:

    def __init__(self, *, filepath: str, use_memory: bool = True, beautify_json: bool = True, **kwargs):
        self._logger = logging.getLogger(self.__class__.__name__)
        raise_for_type(filepath, str)
        filepath = os.path.abspath(filepath.strip())
        if not filepath:
            raise ValueError("not filepath")

        self.filepath = filepath
        self.use_memory = use_memory
        self.beautify_json = beautify_json
        self.saved_json_data: Optional[dict[str, Any]] = None

    def __str__(self) -> str:
        return f"JSONDbFile ({self.filepath}) ({self.count_records()})"

    def __repr__(self) -> str:
        return f"JSONDbFile ({self.filepath}) ({self.count_records()})"

    def __len__(self) -> int:
        return self.count_records()

    @property
    def filename(self) -> str:
        return os.path.split(self.filepath)[1]

    @property
    def dirpath(self) -> str:
        return os.path.split(self.filepath)[0]

    def write_json_data(self, json_data: dict[str, Any]):
        raise_for_type(json_data, dict)

        if self.dirpath and not os.path.exists(self.dirpath):
            os.makedirs(self.dirpath, exist_ok=True)

        with open(self.filepath, mode="w", encoding="utf-8") as f:
            f.write(json.dumps(json_data, ensure_ascii=False, indent=2 if self.beautify_json else None))

        if self.use_memory is True:
            self.saved_json_data = json_data

    def read_json_data(self) -> dict[str, Any]:
        if self.use_memory is True and self.saved_json_data is not None:
            return self.saved_json_data

        if not self.check_exists():
            self.write_json_data({})

        with open(self.filepath, mode="r", encoding="utf-8") as f:
            text_from_file = f.read()
        json_data = json.loads(text_from_file)

        if self.use_memory is True:
            self.saved_json_data = json_data

        return json_data

    def init(self):
        if not self.check_exists():
            self.write_json_data({})

    def reinit(self):
        self.drop()
        self.init()

    def refresh_saved_json_data(self):
        if self.use_memory is True:
            self.saved_json_data = self.read_json_data()

    def get_records(self) -> list[(str, dict[str, Any])]:
        return [(record_id, record) for record_id, record in self.read_json_data().items()]

    def get_record_ids(self) -> list[str]:
        return [record_id for record_id in self.read_json_data().keys()]

    def check_exists(self) -> bool:
        return os.path.exists(self.filepath)

    def check_record_id_exists(self, record_id: str) -> bool:
        return self.get_record(record_id=record_id) is not None

    def count_records(self) -> int:
        return len(self.read_json_data().keys())

    def get_record(self, record_id: str) -> Optional[dict[str, Any]]:
        raise_for_type(record_id, str)
        json_data = self.read_json_data()
        return json_data[record_id] if record_id in json_data.keys() else None

    def generate_record_id(self) -> str:
        record_ids = set(self.get_record_ids())
        res = len(record_ids)
        while str(res) in record_ids:
            res += 1
        return str(res)

    def create_record(
            self,
            record: dict[str, Any],
            record_id: Optional[str] = None,
    ) -> (str, dict[str, Any]):
        if record_id is None:
            record_id = self.generate_record_id()
        raise_for_type(record_id, str)

        json_data = self.read_json_data()
        if record_id in json_data.keys():
            raise KeyError(f"record with record_id={record_id} already exists")

        json_data[record_id] = record
        self.write_json_data(json_data=json_data)

        return record_id, record

    def update_record(
            self,
            *,
            record_id: str,
            record: dict[str, Any]
    ) -> dict[str, Any]:
        raise_for_type(record_id, str)
        raise_for_type(record, dict)

        json_data = self.read_json_data()
        if record_id not in json_data.keys():
            raise ValueError(f"record with record_id='{record_id}' not exists")

        json_data[record_id] = record
        self.write_json_data(json_data)

        return record

    def rm_record(self, record_id: str):
        json_data = self.read_json_data()
        if record_id not in json_data.keys():
            return
        del json_data[record_id]
        self.write_json_data(json_data)

    def rm_records(self, record_ids: list[str]):
        json_data = self.read_json_data()
        for record_id, record in json_data.items():
            if record_id not in record_ids:
                continue
            del json_data[record_id]
        self.write_json_data(json_data)

    def rm_all_records(self):
        self.write_json_data({})

    def copy(self, to_filepath: str):
        self.init()
        shutil.copy(self.filepath, to_filepath)
        return JSONDbFile(filepath=to_filepath, use_memory=self.use_memory, beautify_json=self.beautify_json)

    def drop(self):
        if self.check_exists():
            os.remove(self.filepath)


class BaseJSONDb:

    def __init__(self, json_db_files: list[JSONDbFile] | None = None, **kwargs):
        self._logger = logging.getLogger(self.__class__.__name__)
        if json_db_files is None:
            json_db_files = []
        self.json_db_files: list[JSONDbFile] = json_db_files

    def __str__(self) -> str:
        return f"JSONDbFiles ({len(self.json_db_files)})"

    def __repr__(self) -> str:
        return f"JSONDbFiles ({len(self.json_db_files)})"

    def __len__(self) -> int:
        return len(self.json_db_files)

    def create_json_db_file(
            self, filepath: str, use_memory: bool = False, beautify_json: bool = False, **kwargs
    ) -> JSONDbFile:
        json_db_file = JSONDbFile(filepath=filepath, use_memory=use_memory, beautify_json=beautify_json)
        self.json_db_files.append(json_db_file)
        return json_db_file

    def add_json_db_file(self, json_db_file: JSONDbFile):
        self.json_db_files.append(json_db_file)

    def init(self):
        for file in self.json_db_files:
            file.init()

    def reinit(self):
        for file in self.json_db_files:
            file.reinit()

    def drop(self):
        for file in self.json_db_files:
            file.drop()

    def rm_all_records(self):
        for json_db_file in self.json_db_files:
            json_db_file.rm_all_records()

    def copy_files_to_dir(self, to_dirpath: str) -> Self:
        json_db = BaseJSONDb()
        for json_db_file in self.json_db_files:
            filepath = os.path.join(to_dirpath, json_db_file.filename)
            json_db_file.copy(filepath)
            json_db.create_json_db_file(
                filepath=filepath, use_memory=json_db_file.use_memory, beautify_json=json_db_file.beautify_json
            )
        return json_db


def __example():
    pass


if __name__ == '__main__':
    __example()



# arpakit
# DEPRECATED

import asyncio
import logging
from pathlib import Path
from typing import Optional, Any, Callable

from aiogram import types
from aiogram.exceptions import AiogramError
from aiogram.filters import CommandObject
from pydantic import BaseModel

from arpakitlib.ar_need_type_util import parse_need_type, NeedTypes
from arpakitlib.ar_parse_command_util import BadCommandFormat, parse_command
from arpakitlib.ar_type_util import raise_for_types

_logger = logging.getLogger(__name__)

_logger.warning(f"Module '{Path(__file__).name}' is deprecated; use 'ar_aiogram_as_tg_command_2_util' instead.")


class BadTgCommandFormat(BadCommandFormat):
    pass


class BaseTgCommandParam(BaseModel):
    key: str


class TgCommandFlagParam(BaseTgCommandParam):
    pass


class TgCommandKeyValueParam(BaseTgCommandParam):
    need_type: str
    index: Optional[int] = None
    required: bool = False
    default: Optional[Any] = None

    @property
    def has_index(self) -> bool:
        return False if self.index is None else True


def as_tg_command(
        *params: TgCommandFlagParam | TgCommandKeyValueParam,
        desc: str | None = None,
        passwd_validator: Callable | str | None = None,
        remove_message_after_correct_passwd: bool = True
):
    _PASSWD_KEY = "passwd"
    _HELP_FLAG = "help"

    params = list(params)

    if passwd_validator is not None:
        raise_for_types(passwd_validator, [Callable, str])
        params.append(TgCommandKeyValueParam(key=_PASSWD_KEY, required=True, index=None, need_type=NeedTypes.str_))

    params.append(TgCommandFlagParam(key=_HELP_FLAG))

    _were_keys = set()
    for param in params:
        if param.key in _were_keys:
            raise ValueError(f"key={param.key} is duplicated")
        _were_keys.add(param.key)
    _were_keys.clear()

    _were_indexes = set()
    for param in params:
        if not isinstance(param, TgCommandKeyValueParam) or not param.has_index:
            continue
        if param.index in _were_indexes:
            raise ValueError(f"index={param.index} is duplicated")
        _were_indexes.add(param.index)

    def decorator(handler):

        async def new_handler(*args, **kwargs):
            message: types.Message = args[0]
            if not isinstance(message, types.Message):
                raise TypeError("not isinstance(message, types.Message)")

            command: CommandObject = kwargs["command"]

            try:
                parsed_command = parse_command(command.text)
                kwargs["parsed_command"] = parsed_command

                if (
                        not parsed_command.values_without_key
                        and len(parsed_command.flags) == 1
                        and parsed_command.has_flag(_HELP_FLAG)
                ):
                    text = f"<b>Command</b> /{command.command}"
                    if desc is not None:
                        text += "\n\n" + desc
                    text += "\n\n"

                    if passwd_validator is not None:
                        text += "Passwd is required\n\n"

                    text += "<b>Keys:</b>\n"
                    tg_command_key_value_params = list(filter(lambda p: isinstance(p, TgCommandKeyValueParam), params))
                    if tg_command_key_value_params:
                        for i, _param in enumerate(tg_command_key_value_params):
                            text += f"{i + 1}. <code>{_param.key}</code>"
                            if _param.has_index:
                                text += f", {_param.index}"
                            text += f", {_param.need_type}"
                            if _param.required is True:
                                text += ", <b>required</b>"
                            else:
                                text += ", not required"
                            if _param.default is not None:
                                text += f", <code>{_param.default}</code>"
                            text = text.strip()
                            text += "\n"
                    else:
                        text += "<i>No keys</i>"
                    text = text.strip()
                    text += "\n\n"

                    text += "<b>Flags:</b>\n"
                    tg_command_flag_params = list(filter(lambda p: isinstance(p, TgCommandFlagParam), params))
                    if tg_command_flag_params:
                        for i, _param in enumerate(tg_command_flag_params):
                            text += f"{i + 1}. <code>-{_param.key}</code>\n"
                    else:
                        text += "<i>No flags</i>"
                    text = text.strip()
                    text += "\n\n"

                    text += f"<code>/{command.command} -{_HELP_FLAG}</code>"
                    text += f"\n<code>/{command.command}</code>"

                    await message.answer(text=text.strip())
                    return

                if passwd_validator is not None:
                    passwd_ = parsed_command.get_value_by_key(_PASSWD_KEY)
                    if not passwd_:
                        is_passwd_correct = False
                    elif isinstance(passwd_validator, Callable):
                        is_passwd_correct = passwd_validator(
                            passwd=passwd_, message=message, parsed_command=parsed_command
                        )
                    elif isinstance(passwd_validator, str):
                        is_passwd_correct = (passwd_validator == passwd_)
                    else:
                        raise TypeError("check_passwd is not not Callable and not str")
                    if not is_passwd_correct:
                        await message.answer("Passwd is incorrect")
                        return
                    if remove_message_after_correct_passwd:
                        try:
                            await message.delete()
                        except AiogramError as e:
                            _logger.error(e)
                    try:
                        await message.answer("Passwd is ok")
                    except AiogramError as e:
                        _logger.error(e)

                for _param in params:
                    if isinstance(_param, TgCommandFlagParam):
                        kwargs[_param.key] = parsed_command.has_flag(_param.key)

                    elif isinstance(_param, TgCommandKeyValueParam):
                        if _param.key in kwargs.keys():
                            raise BadTgCommandFormat(f"{_param.key} already in {kwargs.keys()}")

                        value_by_key: Optional[str] = parsed_command.get_value_by_key(_param.key)

                        value_by_index: Optional[str] = None
                        if _param.has_index:
                            value_by_index = parsed_command.get_value_by_index(_param.index)

                        if value_by_key is not None and value_by_index is not None:
                            raise BadTgCommandFormat(
                                f"Value was found by key={_param.key} and index={_param.index}"
                            )

                        value = value_by_key if value_by_key is not None else value_by_index

                        if value is None:
                            if _param.default is not None:
                                value = _param.default
                            elif _param.required is True:
                                raise BadTgCommandFormat(
                                    f"Value (key={_param.key}, index={_param.index}) is required"
                                )
                        else:
                            value = parse_need_type(value=value, need_type=_param.need_type, allow_none=False)

                        kwargs[_param.key] = value

                    else:
                        raise TypeError(f"bad type for param, {type(_param)}")

            except BadCommandFormat as e:
                await message.answer(
                    f"<b>Bad command usage</b> /{command.command}\n\n"

                    "<b>Error</b>\n"
                    f"{e}\n\n"

                    f"Use <code>/{command.command} -{_HELP_FLAG}</code> for getting help info"
                )
                return

            return await handler(*args, **kwargs)

        return new_handler

    return decorator


def __example():
    pass


async def __async_example():
    pass


if __name__ == '__main__':
    __example()
    asyncio.run(__async_example())



# arpakit
from typing import Any, get_type_hints, get_origin, Union, Annotated, get_args

from pydantic import BaseModel, ConfigDict
from sqlalchemy import inspect
from sqlalchemy.orm import ColumnProperty, Mapped


def _define_sqlalchemy_column_mapped_type(type_: Any) -> Any:
    """
    Возвращает тип колонки ИСКЛЮЧИТЕЛЬНО из аннотации поля.
    Разворачивает оболочки:
      - Annotated[T, ...] -> T
      - Mapped[T] -> T
    Если аннотации нет — возвращает Any.
    """

    origin = get_origin(type_)

    if origin is Mapped:
        return get_args(type_)[0] if get_args(type_) else Any

    return type_


def _get_property_name_to_type_from_model_class(
        *,
        model_class: type,
        skip_property_if_cannot_define_type: bool = True,
        exclude_property_names: list[str] | None = None,
        exclude_property_types: list[type] | None = None,
) -> dict[str, Any]:
    """
    Находит все @property в классе и его родителях, и вытаскивает их возвращаемый тип.
    Если тип не удаётся получить — подставляем Any (в зависимости от skip_property_if_cannot_define_type).
    """
    exclude_property_names = set(exclude_property_names or [])
    props: dict[str, Any] = {}

    for cls in reversed(model_class.__mro__):

        if cls is object:
            continue

        for property_name, attr in cls.__dict__.items():
            if not isinstance(attr, property):
                continue

            if property_name in exclude_property_names:
                continue

            try:
                hints = get_type_hints(attr.fget) if attr.fget else {}
                ret_type = hints.get("return", Any)
            except Exception:
                if skip_property_if_cannot_define_type:
                    ret_type = Any
                else:
                    raise

            if exclude_property_types and _type_matches(type_=ret_type, allowed_types=exclude_property_types):
                continue

            props[property_name] = ret_type

    return props


def _type_matches(*, type_: Any, allowed_types: list[type]) -> bool:
    """
    Возвращает True, если тип `tp` соответствует хоть одному из типов в `allowed`.
    - поддерживает Union/Optional (перебирает аргументы),
    - Annotated (смотрит на базовый тип),
    - generics (List[int], Dict[str, Any]) — сравнивает по origin (list, dict, и т.п.).
    """

    origin = get_origin(type_)
    if origin is Union:  # Optional/Union
        return any(_type_matches(type_=arg, allowed_types=allowed_types) for arg in get_args(type_))
    if origin is Annotated:
        return _type_matches(type_=get_args(type_)[0], allowed_types=allowed_types)

    # если это generic, сравним по origin (например, list/dict)
    type_check = origin or type_

    for allowed_type in allowed_types:
        # точное совпадение по объекту типа
        if type_check is allowed_type:
            return True
        # безопасная проверка наследования
        try:
            if isinstance(type_check, type) and isinstance(allowed_type, type) and issubclass(type_check, allowed_type):
                return True
        except Exception:
            pass
    return False


def _get_sqlalchemy_mapped_types(sqlalchemy_model):
    result = {}
    for cls in reversed(sqlalchemy_model.__mro__):
        annotations = getattr(cls, "__annotations__", {})
        for field, annotation in annotations.items():
            if get_origin(annotation) is Mapped:
                result[field] = annotation
    return result


def pydantic_schema_from_sqlalchemy_model(
        sqlalchemy_model: type,
        *,
        model_name: str | None = None,
        base_model: type[BaseModel] = BaseModel,
        include_columns: bool = True,
        exclude_column_names: list[str] | None = None,
        include_properties: bool = False,
        include_property_names: list[str] | None = None,
        include_property_types: list[type] | None = None,
        exclude_property_names: list[str] | None = None,
        exclude_property_types: list[type] | None = None,
        filter_property_prefixes: list[str] | None = None,
        remove_property_prefixes: list[str] | None = None,
        skip_property_if_cannot_define_type: bool = True,
        skip_property_name_if_exists: bool = True,
        model_name_prefix: str | None = "Schema",
        model_name_suffix: str | None = "Schema",
) -> type[BaseModel]:
    """
    Генерирует Pydantic-модель из колонок SQLAlchemy-модели и (опционально) из @property.

    - include_column_defaults: добавлять ли default/server_default у колонок.
    - exclude_column_names: список имён колонок, которые нужно пропустить.

    - include_properties: включать ли свойства (@property). По умолчанию False.
    - include_property_names: whitelist имён свойств (если задан, берём только их).
    - exclude_property_names: blacklist имён свойств (исключаются после whitelist'а).
    """
    mapper = inspect(sqlalchemy_model).mapper

    if not model_name:
        model_name = str(sqlalchemy_model.__name__)
        if model_name_prefix:
            model_name = f"{model_name_prefix}{model_name}"
        if model_name_suffix:
            model_name = f"{model_name}{model_name_suffix}"

    annotations: dict[str, Any] = {}
    attrs: dict[str, Any] = {}

    exclude_column_names = set(exclude_column_names or [])
    include_property_names = set(include_property_names or [])
    exclude_property_names = set(exclude_property_names or [])
    filter_property_prefixes = set(filter_property_prefixes or [])
    remove_property_prefixes = list(set(remove_property_prefixes or []))
    remove_property_prefixes.sort(key=len, reverse=True)

    # 1) Колонки
    if include_columns:

        for column_attr in mapper.column_attrs:
            if not isinstance(column_attr, ColumnProperty):
                continue
            if column_attr.key in exclude_column_names:
                continue
            mapped_type = _get_sqlalchemy_mapped_types(sqlalchemy_model=sqlalchemy_model)[column_attr.key]
            annotations[column_attr.key] = _define_sqlalchemy_column_mapped_type(type_=mapped_type)

    # 2) Свойства (@property)
    if include_properties:
        property_name_to_type = _get_property_name_to_type_from_model_class(
            model_class=sqlalchemy_model,
            skip_property_if_cannot_define_type=skip_property_if_cannot_define_type,
            exclude_property_names=list(exclude_property_names),
            exclude_property_types=exclude_property_types
        )

        # (НОВОЕ) фильтр по типам, если задан
        if include_property_types:
            property_name_to_type = {
                k: v for k, v in property_name_to_type.items()
                if _type_matches(type_=v, allowed_types=include_property_types)
            }

        # Затем исключающие типы (EXCLUDE)
        if exclude_property_types:
            property_name_to_type = {
                k: v
                for k, v in property_name_to_type.items()
                if not _type_matches(type_=v, allowed_types=exclude_property_types)
            }

        # фильтр по префиксам
        if filter_property_prefixes:
            property_name_to_type = {
                property_name: property_type
                for property_name, property_type in property_name_to_type.items()
                if any(property_name.startswith(v) for v in filter_property_prefixes)
            }

        # whitelist по именам
        if include_property_names:
            property_name_to_type = {
                k: v
                for k, v in property_name_to_type.items()
                if k in include_property_names
            }
        else:
            property_name_to_type = dict(property_name_to_type)

        # blacklist по именам
        if exclude_property_names:
            for property_name in list(property_name_to_type.keys()):
                if property_name in exclude_property_names:
                    property_name_to_type.pop(property_name, None)

        # удаление префиксов (без изменений)
        if remove_property_prefixes:
            renamed_property_name_to_type = {}
            for property_name, property_type in list(property_name_to_type.items()):
                new_property_name = property_name
                for prefix in remove_property_prefixes:
                    if new_property_name.startswith(prefix):
                        new_property_name = new_property_name[len(prefix):].strip()
                        break
                if new_property_name != property_name:
                    property_name_to_type.pop(property_name)
                    # избегаем коллизий
                    if (
                            new_property_name not in annotations
                            and new_property_name not in property_name_to_type
                            and new_property_name not in renamed_property_name_to_type
                    ):
                        renamed_property_name_to_type[new_property_name] = property_type
                    else:
                        if not skip_property_name_if_exists:
                            raise ValueError(
                                f"Property name '{property_name}' after removing prefix "
                                f"conflicts with existing name '{new_property_name}'"
                            )
            property_name_to_type.update(renamed_property_name_to_type)

        # добавляем (колонки в приоритете)
        for property_name, property_type in list(property_name_to_type.items()):
            if property_name in annotations:
                continue
            annotations[property_name] = property_type

    attrs["__annotations__"] = annotations

    if "model_config" not in attrs:
        attrs["model_config"] = ConfigDict(extra="ignore", arbitrary_types_allowed=True, from_attributes=True)

    return type(model_name, (base_model,), attrs)



# arpakit


import logging


async def log_async_func_if_error(
        async_func,
        logger: logging.Logger | None = None,
        logger_name: str | None = None,
        **kwargs
):
    if logger is None:
        logger = logging.getLogger()
    if logger is None and logger_name is not None:
        logger = logging.getLogger(logger_name)

    try:
        await async_func(**kwargs)
    except Exception as exception:
        logger.error(
            f"error in async_func, {async_func.__name__=}",
            exc_info=exception,
            extra={
                "log_async_func_if_error": True,
                "async_func_name": async_func.__name__
            }
        )



# arpakit


class CollectingSubclassesMeta(type):
    """
    Метакласс для автоматического сбора всех наследников в поле ALL_SUBCLASSES.
    """

    def __init__(cls, name, bases, dct, **kwargs):
        super().__init__(name, bases, dct, **kwargs)
        if not hasattr(cls, "all_subclasses"):
            cls.all_subclasses = []
        elif bases:
            cls.all_subclasses.append(cls)


def create_combined_meta(*metas):
    """
    Создает объединённый метакласс для устранения конфликтов.
    """

    class CombinedMeta(*metas):
        pass

    return CombinedMeta



import traceback


def exception_to_traceback_str(exception: BaseException) -> str:
    try:
        return "".join(traceback.format_exception(type(exception), exception, exception.__traceback__))
    except Exception as exception_:
        return f"Failed to format exception to traceback str {exception!r}: {exception_!r}"


def __example():
    pass


if __name__ == '__main__':
    __example()



# arpakit

from __future__ import annotations

import asyncio
import io
import logging
from datetime import timedelta
from typing import Any

import asyncssh
import paramiko
from arpakitlib.ar_json_util import transfer_data_to_json_str
from pydantic import BaseModel, ConfigDict


class BaseSSHException(Exception):
    pass


class ConnectionSSHException(BaseSSHException):
    def __init__(self, ssh_runner: SSHRunner, base_exception: Exception | None = None,
                 error_message: str | None = None):
        self.ssh_runner = ssh_runner
        self.base_exception = base_exception
        self.error_message = error_message

    def format_message(self) -> str:
        parts = [
            f"{self.__class__.__name__}",
            f"{self.ssh_runner.username}@{self.ssh_runner.hostname}:{self.ssh_runner.port}",
            f"{type(self.base_exception)=}",
            f"{self.base_exception=}"
        ]
        return ', '.join(parts)

    def __str__(self) -> str:
        return self.format_message()

    def __repr__(self) -> str:
        return self.format_message()


class ErrorInRunSSHException(BaseSSHException):
    def __init__(self, ssh_runner: SSHRunner, base_exception: Exception | None = None, message: str | None = None):
        self.ssh_runner = ssh_runner
        self.base_exception = base_exception
        self.message = message
        super().__init__(self.format_message())

    def format_message(self) -> str:
        parts = [
            f"{self.__class__.__name__}",
            f"{self.ssh_runner.username}@{self.ssh_runner.hostname}:{self.ssh_runner.port}",
        ]
        if self.base_exception is not None:
            parts.append(f"{type(self.base_exception)=}")
            parts.append(f"{self.base_exception=}")
        if self.message is not None:
            parts.append(f"{self.message=}")
        return ', '.join(parts)

    def __str__(self) -> str:
        return self.format_message()

    def __repr__(self) -> str:
        return self.format_message()


class SSHRunResultHasErrorSSHException(BaseSSHException):
    def __init__(self, ssh_run_result: SSHRunResult, message: str | None = None):
        self.ssh_run_result = ssh_run_result
        self.message = message
        super().__init__(self.format_message())

    def format_message(self) -> str:
        parts = [
            f"{self.__class__.__name__}",
            f"{self.ssh_run_result.ssh_runner.username}@{self.ssh_run_result.ssh_runner.hostname}:{self.ssh_run_result.ssh_runner.port}",
            f"{str(self.ssh_run_result.return_code)=}",
            f"{str(self.ssh_run_result.err)=}",
        ]
        if self.message is not None:
            parts.append(f"{self.message=}")
        return ', '.join(parts)

    def __str__(self) -> str:
        return self.format_message()

    def __repr__(self) -> str:
        return self.format_message()


class SSHRunResult(BaseModel):
    out: str
    err: str
    return_code: int | None = None
    ssh_runner: SSHRunner

    model_config = ConfigDict(extra="forbid", arbitrary_types_allowed=True, from_attributes=True)

    def simple_dict(self) -> dict[str, Any]:
        return {
            "out": self.out,
            "err": self.err,
            "return_code": self.return_code,
            "has_bad_return_code": self.has_bad_return_code,
            "has_err": self.has_err,
            "has_out": self.has_out,
            "username": self.ssh_runner.username,
            "hostname": self.ssh_runner.hostname,
            "port": self.ssh_runner.port,
        }

    def simple_json(self) -> str:
        return transfer_data_to_json_str(
            self.simple_dict(),
            beautify=True,
            fast=False
        )

    def __repr__(self) -> str:
        return self.simple_json()

    def __str__(self) -> str:
        return self.simple_json()

    @property
    def has_bad_return_code(self) -> bool:
        if self.return_code is None:
            return False
        return self.return_code != 0

    @property
    def has_err(self) -> bool:
        if self.err:
            return True
        return False

    @property
    def has_out(self) -> bool:
        if self.out:
            return True
        return False

    def raise_for_bad_return_code(self):
        if self.has_bad_return_code:
            raise SSHRunResultHasErrorSSHException(ssh_run_result=self)

    def raise_for_err(self):
        if self.has_err:
            raise SSHRunResultHasErrorSSHException(ssh_run_result=self)


class SSHRunner:

    def __init__(
            self,
            *,
            username: str = "root",
            hostname: str,  # ipv4, ipv6, domain
            port: int = 22,
            password: str | None = None,
            private_key: str | None = None,
            base_timeout: float | None = None,
            check_if_already_connected: bool | None = True,
            auto_close_after_run: bool | None = False
    ):
        self.username = username
        self.hostname = hostname
        self.port = port
        self.password = password
        self.private_key = private_key

        if base_timeout is None:
            base_timeout = timedelta(seconds=10).total_seconds()
        self.base_timeout = base_timeout

        if check_if_already_connected is None:
            check_if_already_connected = True
        self.check_if_already_connected = check_if_already_connected

        if auto_close_after_run is None:
            auto_close_after_run = True
        self.auto_close_after_run = auto_close_after_run

        self._logger = logging.getLogger(
            f"{logging.getLogger(self.__class__.__name__)} - {self.username}@{self.hostname}:{self.port}"
        )

        self.async_conn: asyncssh.SSHClientConnection | None = None

        self.sync_client: paramiko.SSHClient | None = None

    """SYNC"""

    def __enter__(self) -> SSHRunner:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.sync_close()
        return False

    def sync_connect(
            self,
            *,
            common_timeout: float | None = None,
            connect_kwargs: dict[str, Any] | None = None,
            check_if_already_connected: bool | None = None
    ) -> SSHRunner:
        if check_if_already_connected is None:
            check_if_already_connected = self.check_if_already_connected

        if check_if_already_connected and self.sync_client is not None:
            self._logger.info("already connected")
            return self

        self.sync_close()

        if connect_kwargs is None:
            connect_kwargs = {}
        if common_timeout is None:
            common_timeout = self.base_timeout

        connect_kwargs["hostname"] = self.hostname
        connect_kwargs["username"] = self.username
        connect_kwargs["password"] = self.password
        connect_kwargs["port"] = self.port

        if connect_kwargs.get("timeout") is None:
            connect_kwargs["timeout"] = common_timeout
        if connect_kwargs.get("auth_timeout") is None:
            connect_kwargs["auth_timeout"] = common_timeout
        if connect_kwargs.get("banner_timeout") is None:
            connect_kwargs["banner_timeout"] = common_timeout
        if connect_kwargs.get("channel_timeout") is None:
            connect_kwargs["channel_timeout"] = common_timeout

        connect_kwargs.setdefault("allow_agent", False)
        connect_kwargs.setdefault("look_for_keys", False)

        if self.private_key:
            private_key_to_set = None
            try:
                private_key_to_set = paramiko.PKey.from_private_key(io.StringIO(self.private_key))
            except Exception:
                pass
            try:
                private_key_to_set = paramiko.Ed25519Key.from_private_key(file_obj=io.StringIO(self.private_key))
            except Exception:
                pass
            try:
                private_key_to_set = paramiko.RSAKey.from_private_key(file_obj=io.StringIO(self.private_key))
            except Exception:
                pass
            if private_key_to_set is not None:
                connect_kwargs["pkey"] = private_key_to_set
            else:
                self._logger.warning(f"incorrect private key, {self.private_key=}")

        self._logger.info("connecting")

        self.sync_client = paramiko.SSHClient()
        self.sync_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

        try:
            self.sync_client.connect(**connect_kwargs)
        except Exception as exception:
            self.sync_close()
            raise ConnectionSSHException(ssh_runner=self, base_exception=exception, error_message=f"{exception=}")

        self._logger.info("connected")

        return self

    def sync_check_connection(self):
        self.sync_connect()

    def sync_is_conn_good(self) -> bool:
        try:
            self.sync_check_connection()
        except ConnectionSSHException:
            return False
        except Exception:
            return False
        return True

    def sync_run(
            self,
            command: str,
            *,
            timeout: float | None = timedelta(seconds=10).total_seconds(),
            raise_for_bad_return_code: bool = True,
            stdin_data: str | bytes | None = None,
    ) -> SSHRunResult:
        if not command or not command.strip():
            raise ValueError("command must be a non-empty string")

        if timeout is None:
            timeout = self.base_timeout

        self.sync_connect()

        self._logger.info(command)

        try:
            stdin, stdout, stderr = self.sync_client.exec_command(
                command=command,
                timeout=timeout
            )

            if stdin_data is not None:
                if isinstance(stdin_data, str):
                    stdin_data = stdin_data.encode()
                stdin.write(stdin_data)
                stdin.flush()
                stdin.channel.shutdown_write()
                stdin.close()

            out = stdout.read().decode(errors="replace")
            err = stderr.read().decode(errors="replace")
            return_code = stdout.channel.recv_exit_status()
        except Exception as exception:
            if self.auto_close_after_run:
                self.sync_close()
            raise ErrorInRunSSHException(ssh_runner=self, base_exception=exception)

        if self.auto_close_after_run:
            self.sync_close()

        ssh_run_result = SSHRunResult(
            out=out,
            err=err,
            return_code=return_code,
            ssh_runner=self
        )

        if raise_for_bad_return_code:
            ssh_run_result.raise_for_bad_return_code()

        return ssh_run_result

    def sync_close(self):
        if self.sync_client is not None:
            self.sync_client.close()
            self.sync_client = None
        self._logger.info("sync_client was closed")

    """ASYNC"""

    async def __aenter__(self) -> SSHRunner:
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.async_close()
        return False

    async def async_connect(
            self,
            *,
            common_timeout: float | None = None,
            connect_kwargs: dict[str, Any] | None = None,
            check_if_already_connected: bool | None = None
    ) -> SSHRunner:
        if check_if_already_connected is None:
            check_if_already_connected = self.check_if_already_connected

        if check_if_already_connected and self.async_conn is not None:
            self._logger.info("already connected")
            return self

        await self.async_close()

        if connect_kwargs is None:
            connect_kwargs = {}
        if common_timeout is None:
            common_timeout = self.base_timeout

        connect_kwargs["host"] = self.hostname
        connect_kwargs["username"] = self.username
        connect_kwargs["password"] = self.password
        connect_kwargs["port"] = self.port

        if connect_kwargs.get("connect_timeout") is None:
            connect_kwargs["connect_timeout"] = common_timeout
        if connect_kwargs.get("login_timeout") is None:
            connect_kwargs["login_timeout"] = common_timeout

        connect_kwargs["known_hosts"] = None

        self._logger.info("connecting")

        try:
            self.async_conn = await asyncssh.connect(**connect_kwargs)
        except Exception as exception:
            await self.async_close()
            raise ConnectionSSHException(ssh_runner=self, base_exception=exception)

        self._logger.info("connected")

        return self

    async def async_check_connection(self):
        await self.async_connect()

    async def async_is_conn_good(self) -> bool:
        try:
            await self.async_check_connection()
        except ConnectionSSHException:
            return False
        return True

    async def async_run(
            self,
            command: str,
            *,
            timeout: float | None = timedelta(seconds=10).total_seconds(),
            raise_for_bad_return_code: bool = True
    ) -> SSHRunResult:
        if not command or not command.strip():
            raise ValueError("command must be a non-empty string")

        if timeout is None:
            timeout = self.base_timeout

        await self.async_connect()

        self._logger.info(command)

        try:
            result: asyncssh.SSHCompletedProcess = await self.async_conn.run(
                command,
                check=False,
                timeout=timeout
            )
        except Exception as exception:
            if self.auto_close_after_run:
                await self.async_close()
            raise ErrorInRunSSHException(ssh_runner=self, base_exception=exception)

        if self.auto_close_after_run:
            await self.async_close()

        ssh_run_result = SSHRunResult(
            out=result.stdout,
            err=result.stderr,
            return_code=result.returncode,
            ssh_runner=self
        )

        if raise_for_bad_return_code:
            ssh_run_result.raise_for_bad_return_code()

        return ssh_run_result

    async def async_close(self):
        if self.async_conn is not None:
            self.async_conn.close()
            self.async_conn = None
        self._logger.info("sync_client was closed")


def __example():
    pass


async def __async_example():
    pass


if __name__ == '__main__':
    __example()
    asyncio.run(__async_example())



# arpakit

import logging
import os

from arpakitlib.ar_str_util import raise_if_string_blank
from arpakitlib.ar_type_util import raise_for_type

_logger = logging.getLogger(__name__)


def init_arpakit_project_template(
        *,
        version: str = "1",
        project_dirpath: str = "./",
        overwrite_if_exists: bool = False,
        ignore_paths_startswith: list[str] | str | None = None,
        only_paths_startswith: list[str] | str | None = None,
):
    raise_for_type(project_dirpath, str)
    raise_if_string_blank(project_dirpath)

    raise_for_type(overwrite_if_exists, bool)

    if isinstance(ignore_paths_startswith, str):
        ignore_paths_startswith = [ignore_paths_startswith]
    if ignore_paths_startswith is None:
        ignore_paths_startswith = []
    raise_for_type(ignore_paths_startswith, list)

    if isinstance(only_paths_startswith, str):
        only_paths_startswith = [only_paths_startswith]
    if only_paths_startswith is None:
        only_paths_startswith = []
    raise_for_type(only_paths_startswith, list)

    def _generate_filepath_to_content() -> dict[str, bytes]:
        arpakit_project_template_dirpath = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), f"_arpakit_project_template_v_{version}"
        )
        if not os.path.exists(arpakit_project_template_dirpath):
            raise Exception(f"not os.path.exists({arpakit_project_template_dirpath})")
        res = {}
        for root, dirs, files in os.walk(arpakit_project_template_dirpath):
            dirs[:] = [d for d in dirs if d != '__pycache__']
            for file in files:
                rel_path = os.path.relpath(os.path.join(root, file), arpakit_project_template_dirpath)
                if (
                        ignore_paths_startswith
                        and
                        any(rel_path.startswith(ignore_path) for ignore_path in ignore_paths_startswith)
                ):
                    _logger.info(f"ignoring file: {rel_path}")
                    continue
                if (
                        only_paths_startswith
                        and
                        not any(rel_path.startswith(only_path) for only_path in only_paths_startswith)
                ):
                    _logger.info(f"ignoring file: {rel_path}")
                    continue
                with open(os.path.join(root, file), "rb") as _file:
                    _content = _file.read()
                res[rel_path] = _content
        return res

    filepath_to_content = _generate_filepath_to_content()

    if not os.path.exists(project_dirpath):
        os.makedirs(project_dirpath)

    for filepath, content in filepath_to_content.items():
        full_filepath = os.path.join(project_dirpath, filepath)

        if os.path.exists(full_filepath) and not overwrite_if_exists:
            _logger.info(f"file exists and overwrite_if_exists is False, skipping: {full_filepath}")
            continue

        _logger.info(f"creating file: {full_filepath}")
        os.makedirs(os.path.dirname(full_filepath), exist_ok=True)
        with open(full_filepath, "wb") as file_:
            file_.write(content)
        _logger.info(f"file created: {full_filepath}")



from __future__ import annotations

from typing import Optional

import sqlalchemy
from sqlalchemy.engine import Engine
from sqlalchemy.orm import DeclarativeBase


def _quote_ident_pg(ident: str) -> str:
    # простое кавычирование для PG (экранируем двойные кавычки)
    return '"' + ident.replace('"', '""') + '"'


def _quote_ident_mysql(ident: str) -> str:
    # простое кавычирование для MySQL/MariaDB (экранируем обратные апострофы)
    return '`' + ident.replace('`', '``') + '`'


def _qualified_name(
        schema: Optional[str], table: str, dialect_name: str
) -> str:
    if dialect_name == "postgresql":
        q = _quote_ident_pg
    elif dialect_name in {"mysql", "mariadb"}:
        q = _quote_ident_mysql
    else:
        # по умолчанию без кавычек — но мы не должны сюда попасть,
        # поскольку ниже ограничиваем поддержанные диалекты.
        q = lambda x: x

    if schema:
        return f"{q(schema)}.{q(table)}"
    return q(table)


def drop_sqlalchemy_check_constraints(*, base_: type[DeclarativeBase], engine: Engine) -> None:
    """
    Удаляет ВСЕ существующие в БД CHECK-ограничения для таблиц из base_.metadata,
    выполняя прямые SQL-запросы (без SQLAlchemy DDL-объектов).

    Поддерживаемые диалекты:
      - PostgreSQL: ALTER TABLE <sch>.<tbl> DROP CONSTRAINT <name>
      - MySQL 8.0.16+/MariaDB 10.2+: ALTER TABLE <sch>.<tbl> DROP CHECK <name>

    Для SQLite кидает NotImplementedError.
    """

    if engine.dialect.name == "sqlite":
        raise NotImplementedError(
            "Удаление CHECK-констреинтов прямыми запросами для SQLite не поддержано: "
            "в большинстве версий требуется пересоздание таблицы."
        )
    if engine.dialect.name not in {"postgresql", "mysql", "mariadb"}:
        raise NotImplementedError(f"Пока не реализовано для диалекта: {engine.dialect.name}")

    with engine.begin() as conn:
        conn_inspection_ = sqlalchemy.inspect(conn)

        for table in base_.metadata.tables.values():
            if not conn_inspection_.has_table(table.name, schema=table.schema):
                continue

            fqtn = _qualified_name(table.schema, table.name, engine.dialect.name)

            # берём ВСЕ реальные CHECK-и из базы
            checks = conn_inspection_.get_check_constraints(table.name, schema=table.schema)

            for check_ in checks:
                name = check_.get("name")
                if not name:
                    # На всякий случай: без имени нам нечего дропать безопасно
                    # (в PG/ MySQL синтаксис требует имя).
                    continue

                if engine.dialect.name == "postgresql":
                    ident = _quote_ident_pg(name)
                    sql = f"ALTER TABLE {fqtn} DROP CONSTRAINT {ident}"
                else:  # mysql / mariadb
                    ident = _quote_ident_mysql(name)
                    # В MySQL/MariaDB именно DROP CHECK <name>
                    sql = f"ALTER TABLE {fqtn} DROP CHECK {ident}"

                # выполняем прямой SQL-запрос
                conn.exec_driver_sql(sql)



from random import randint


def generate_simple_code(
        *,
        amount: int = 5
) -> str:
    alphabet: list = list("JZSDQWRLGFZ" + "123456789")
    return "".join(alphabet[randint(0, len(alphabet) - 1)] for _ in range(amount))


def __example():
    pass


if __name__ == '__main__':
    __example()



from __future__ import annotations

from typing import Any

from sqladmin import ModelView
from sqlalchemy.orm.attributes import InstrumentedAttribute


def get_string_info_from_model_view(class_: type[ModelView]):
    lines = [f"ModelViews: {len(class_.__subclasses__())}"]
    for i, cls in enumerate(class_.__subclasses__()):
        if hasattr(cls, "__name__"):
            lines.append(f"{i + 1}. ModelView: {cls.__name__}, (columns={len(cls.column_list)})")
            for v in cls.column_list:
                v: str | InstrumentedAttribute | Any
                if isinstance(v, str):
                    lines.append(f"   - {v}")
                elif isinstance(v, InstrumentedAttribute):
                    lines.append(f"   - {v}")
                else:
                    raise TypeError(f"unknown type, {type(v)=}")

    return "\n".join(lines)



# arpakit
import hashlib
import json
import os
import pathlib
from importlib.util import spec_from_file_location, module_from_spec
from typing import NamedTuple, Any, Iterator


class ArpakitLibModule(NamedTuple):
    module_name: str
    module_content: str
    module_hash: str
    module_has_error: bool
    module_exception: Exception | None
    filename: str
    filepath: str

    def simple_dict(self) -> dict[str, Any]:
        return {
            "module_name": self.module_name,
            "module_content": self.module_content,
            "module_hash": self.module_hash,
            "module_has_error": self.module_has_error,
            "module_exception": self.module_exception,
            "filename": self.filename,
            "filepath": self.filepath,
        }


class ArpakitLibModules(NamedTuple):
    arpakit_lib_modules: list[ArpakitLibModule]

    def __len__(self):
        return len(self.arpakit_lib_modules)

    def __repr__(self):
        return f"ArpakitLibModules (len={len(self.arpakit_lib_modules)})"

    def __iter__(self) -> Iterator[ArpakitLibModule]:
        for arpakit_lib_module in self.arpakit_lib_modules:
            yield arpakit_lib_module

    def __hash__(self) -> str:
        return self.modules_hash()

    def modules_hash(self) -> str:
        return hashlib.sha256(
            json.dumps(self.module_name_to_module_hash()).encode("utf-8")
        ).hexdigest()

    def simple_dict(self) -> dict[str, list[dict[str, Any]]]:
        return {
            "arpakit_lib_modules": [arpakit_lib_module.simple_dict() for arpakit_lib_module in self.arpakit_lib_modules]
        }

    def module_name_to_module_simple_dict(self) -> dict[str, dict]:
        return {module.module_name: module.simple_dict() for module in self.arpakit_lib_modules}

    def module_name_to_has_errors(self) -> dict[str, dict[str, Any]]:
        return {
            module.module_name: {
                "module_has_errors": module.module_has_error,
            } for module in self.arpakit_lib_modules
        }

    def module_names_who_has_errors(self) -> list[str]:
        return [
            arpakit_lib_module.module_name
            for arpakit_lib_module in self.arpakit_lib_modules
            if arpakit_lib_module.module_has_error
        ]

    def have_modules_with_error(self) -> bool:
        for arpakit_lib_module in self.arpakit_lib_modules:
            if arpakit_lib_module.module_has_error:
                return True
        return False

    def module_name_to_module_exception(self, *, filter_module_has_error: bool = False) -> dict[str, Exception]:
        if filter_module_has_error:
            return {
                module.module_name: module.module_exception
                for module in self.arpakit_lib_modules if module.module_has_error
            }
        else:
            return {
                module.module_name: module.module_exception
                for module in self.arpakit_lib_modules
            }

    def module_name_to_module_hash(self) -> dict[str, str]:
        return {module.module_name: module.module_hash for module in self.arpakit_lib_modules}

    def module_name_to_module_content(self) -> dict[str, str]:
        return {module.module_name: module.module_content for module in self.arpakit_lib_modules}


def get_arpakitlib_modules() -> ArpakitLibModules:
    base_dirpath: str = str(pathlib.Path(__file__).parent)

    filenames: list[str] = os.listdir(base_dirpath)
    filenames.sort()

    arpakit_lib_modules = ArpakitLibModules(arpakit_lib_modules=[])

    for filename in filenames:
        if not filename.endswith(".py") or filename == "__init__.py":
            continue
        module_name = filename.replace(".py", "")
        try:
            spec = spec_from_file_location(module_name, os.path.join(base_dirpath, filename))
            module = module_from_spec(spec)
            spec.loader.exec_module(module)
            module_has_error = False
            module_exception = None
        except Exception as error:
            module_has_error = True
            module_exception = error
        if module_name in [
            arpakit_lib_module.module_name for arpakit_lib_module in arpakit_lib_modules.arpakit_lib_modules
        ]:
            raise KeyError(f"module_name {module_name} is duplicated")
        module_content = open(file=os.path.join(base_dirpath, filename), mode="r").read().strip()
        module_hash = hashlib.sha256(module_content.encode('utf-8')).hexdigest()
        arpakit_lib_modules.arpakit_lib_modules.append(ArpakitLibModule(
            module_name=module_name,
            module_content=module_content,
            module_hash=module_hash,
            module_has_error=module_has_error,
            module_exception=module_exception,
            filename=filename,
            filepath=os.path.join(base_dirpath, filename)
        ))

    return arpakit_lib_modules


def __example():
    print(get_arpakitlib_modules())


if __name__ == '__main__':
    __example()



# arpakit

import hashlib


def hash_string(string: str) -> str:
    return hashlib.sha256(string.encode()).hexdigest()


def check_string_hash(string: str, string_hash: str) -> bool:
    return hash_string(string) == string_hash


def __example():
    pass


if __name__ == '__main__':
    __example()



# arpakit

import asyncio
import logging
import multiprocessing
import threading
from abc import ABC
from datetime import timedelta
from random import randint
from typing import Any
from uuid import uuid4

from arpakitlib.ar_enumeration_util import Enumeration
from arpakitlib.ar_func_util import is_async_func, is_sync_func, is_coroutine
from arpakitlib.ar_sleep_util import sync_safe_sleep, async_safe_sleep


class BaseWorker(ABC):
    def __init__(
            self,
            *,
            timeout_after_run: timedelta = timedelta(seconds=0.3),
            timeout_after_error_in_run: timedelta = timedelta(seconds=1),
            startup_funcs: list[Any] | None = None,
            worker_name: str | None = None,
            data: dict[str, Any] | None = None,
            timeout_before_safe_run: timedelta | None = None,
            **kwargs
    ):
        self.timeout_after_run = timeout_after_run
        self.timeout_after_error_in_run = timeout_after_error_in_run

        if startup_funcs is None:
            startup_funcs = []
        self.startup_funcs = startup_funcs

        if worker_name is None:
            worker_name = self.__class__.__name__
        self.worker_name = worker_name
        self.worker_id = f"{str(uuid4()).replace('-', '')}_{randint(1000, 99999)}"

        if data is None:
            data = {}
        self.data = data

        self.timeout_before_safe_run = timeout_before_safe_run

        self._logger = logging.getLogger(self.worker_fullname)

    @property
    def worker_fullname(self) -> str:
        return f"{self.worker_name}_{self.worker_id}"

    def sync_run_startup_funcs(self):
        self._logger.info("start")
        for startup_func in self.startup_funcs:
            if is_async_func(startup_func):
                asyncio.run(startup_func())
            elif is_coroutine(startup_func):
                async def __func():
                    await startup_func

                asyncio.run(__func())
            elif is_sync_func(startup_func):
                startup_func()
            else:
                raise TypeError("unknown startup_func type")
        self._logger.info("finish")

    def sync_on_startup(self):
        self.sync_run_startup_funcs()

    def sync_run(self):
        pass

    def sync_on_error(self, exception: Exception, **kwargs):
        pass

    def sync_safe_run(self):
        self._logger.info("start")
        if self.timeout_before_safe_run is not None:
            sync_safe_sleep(self.timeout_before_safe_run)
        try:
            self.sync_on_startup()
        except Exception as exception:
            self._logger.error("exception in sync_on_startup", exc_info=exception)
            raise
        while True:
            try:
                self.sync_run()
            except Exception as exception:
                self._logger.error("exception in sync_run", exc_info=exception)
                try:
                    self.sync_on_error(exception=exception)
                except Exception as exception_:
                    self._logger.error("exception in sync_on_error", exc_info=exception_)
                    raise
                sync_safe_sleep(self.timeout_after_error_in_run)
            sync_safe_sleep(self.timeout_after_run)

    async def async_run_startup_funcs(self):
        self._logger.info("start")
        for startup_func in self.startup_funcs:
            if is_async_func(startup_func):
                await startup_func()
            elif is_coroutine(startup_func):
                await startup_func
            elif is_sync_func(startup_func):
                startup_func()
            else:
                raise TypeError("unknown startup_func type")
        self._logger.info("finish")

    async def async_on_startup(self):
        await self.async_run_startup_funcs()

    async def async_run(self):
        pass

    async def async_on_error(self, exception: Exception, **kwargs):
        pass

    async def async_safe_run(self):
        self._logger.info("start async_safe_run")
        if self.timeout_before_safe_run is not None:
            await async_safe_sleep(self.timeout_before_safe_run)
        try:
            await self.async_on_startup()
        except Exception as exception:
            self._logger.error("exception in async_on_startup", exc_info=exception)
            raise
        while True:
            try:
                await self.async_run()
            except Exception as exception:
                self._logger.error("exception in async_run", exc_info=exception)
                try:
                    await self.async_on_error(exception=exception)
                except Exception as exception_:
                    self._logger.error("exception in async_on_error", exc_info=exception_)
                    raise
                await async_safe_sleep(self.timeout_after_error_in_run)
            await async_safe_sleep(self.timeout_after_run)


class SafeRunInBackgroundModes(Enumeration):
    async_task = "async_task"
    thread = "thread"
    process = "process"


def safe_run_worker_in_background(*, worker: BaseWorker, mode: str) -> (
        asyncio.Task | threading.Thread | multiprocessing.Process
):
    if mode == SafeRunInBackgroundModes.async_task:
        res: asyncio.Task = asyncio.create_task(worker.async_safe_run())
    elif mode == SafeRunInBackgroundModes.thread:
        res: threading.Thread = threading.Thread(
            target=worker.sync_safe_run,
            daemon=True
        )
        res.start()
    elif mode == SafeRunInBackgroundModes.process:
        res: multiprocessing.Process = multiprocessing.Process(
            target=worker.sync_safe_run,
            daemon=True
        )
        res.start()
    else:
        raise ValueError(f"unknown safe_run_mode={mode}")
    return res


def safe_run_workers_in_background(
        *, workers: list[BaseWorker], mode: str
) -> list[asyncio.Task] | list[threading.Thread] | list[multiprocessing.Process]:
    res = []
    for worker in workers:
        res.append(safe_run_worker_in_background(worker=worker, mode=mode))
    return res


async def a():
    pass


def __example():
    pass


async def __async_example():
    pass


if __name__ == '__main__':
    __example()
    asyncio.run(__async_example())



# arpakit
import asyncio
import logging
from datetime import timedelta, datetime
from typing import Any, Collection
from uuid import uuid4

import sqlalchemy
from sqlalchemy import create_engine, QueuePool, text, func, inspect, AsyncAdaptedQueuePool
from sqlalchemy.ext.asyncio import create_async_engine, async_sessionmaker, AsyncSession
from sqlalchemy.orm import sessionmaker, DeclarativeBase
from sqlalchemy.orm.session import Session

from arpakitlib.ar_datetime_util import now_utc_dt


class BaseDBM(DeclarativeBase):
    __abstract__ = True
    __table_args__ = {"extend_existing": True}

    @classmethod
    def get_column_names_(
            cls,
            *,
            include_pk: bool = True,
            exclude_names: list[str] | None = None,
            exclude_if_have_foreign_keys: bool = False,
            filter_prefixes: list[str] | None = None
    ) -> list[str]:
        if exclude_names is None:
            exclude_names = []
        res = []
        for c in inspect(cls).columns:
            if not include_pk and c.primary_key:
                continue
            if c.key in exclude_names:
                continue
            if exclude_if_have_foreign_keys and c.foreign_keys:
                continue
            if filter_prefixes is not None and not any(
                    c.key.startswith(prefix) for prefix in filter_prefixes
            ):
                continue
            res.append(c.key)
        return res

    @classmethod
    def get_relationship_names_(
            cls,
            *,
            exclude_one_to_many: bool = False,
            exclude_many_to_one: bool = False,
            exclude_many_to_many: bool = False,
            exclude_viewonly: bool = False,
            exclude_names: list[str] | None = None
    ) -> list[str]:
        if exclude_names is None:
            exclude_names = []
        res = []
        for r in inspect(cls).relationships:
            if r.key in exclude_names:
                continue
            if exclude_viewonly and r.viewonly:
                continue
            if exclude_one_to_many and r.direction is sqlalchemy.orm.ONETOMANY:
                continue
            if exclude_many_to_one and r.direction is sqlalchemy.orm.MANYTOONE:
                continue
            if exclude_many_to_many and r.direction is sqlalchemy.orm.MANYTOMANY:
                continue
            res.append(r.key)
        return res

    @classmethod
    def get_column_and_relationship_names_(
            cls,
            *,

            include_column_names: bool = True,
            include_column_pk: bool = True,
            exclude_column_names: list[str] | None = None,
            exclude_column_names_if_have_foreign_keys: bool = False,

            include_relationship_names: bool = True,
            exclude_relationship_one_to_many: bool = False,
            exclude_relationship_many_to_one: bool = False,
            exclude_relationship_many_to_many: bool = False,
            exclude_relationship_names: list[str] | None = None,
    ) -> list[str]:
        res = []
        if include_column_names:
            res += cls.get_column_names_(
                include_pk=include_column_pk,
                exclude_names=exclude_column_names,
                exclude_if_have_foreign_keys=exclude_column_names_if_have_foreign_keys
            )
        if include_relationship_names:
            res += cls.get_relationship_names_(
                exclude_names=exclude_relationship_names,
                exclude_one_to_many=exclude_relationship_one_to_many,
                exclude_many_to_one=exclude_relationship_many_to_one,
                exclude_many_to_many=exclude_relationship_many_to_many,
            )
        return res

    @classmethod
    def get_sd_property_names(
            cls,
            *,
            prefixes: list[str] | None = None,
            remove_prefixes: list[str] | None = None,
            exclude_property_names: list[str] | None = None,
            exclude_property_names_after_remove_prefix: list[str] | None = None,
    ) -> list[str]:
        """
        Возвращает список имён sd-свойств (property), начинающихся с указанных префиксов.

        :param prefixes: Список префиксов, по которым определяются sd-свойства.
        :param remove_prefixes: Префиксы, которые нужно удалить из имён перед возвратом.
                                Если не указаны, префиксы не удаляются.
        :param exclude_property_names: Список имён (до удаления префикса), которые нужно исключить.
        :param exclude_property_names_after_remove_prefix: Список имён (после удаления префиксов),
                                которые нужно исключить.
        :return: Список имён свойств (без дубликатов), упорядоченных от потомков к базам.
        """
        exclude_property_names = set(exclude_property_names or [])
        exclude_property_names_after_remove_prefix = set(exclude_property_names_after_remove_prefix or [])
        res: list[str] = []

        # идём от потомка к базовым классам — переопределения потомка "побеждают"
        for c in cls.__mro__:
            if c is object:
                continue

            for name, attr in c.__dict__.items():
                if not isinstance(attr, property):
                    continue

                # если заданы prefixes, фильтруем по ним
                if prefixes and not any(name.startswith(p) for p in prefixes):
                    continue

                if name in exclude_property_names or name in res:
                    continue

                # удаляем префикс (если требуется)
                clean_name = name
                if remove_prefixes:
                    for p in remove_prefixes:
                        if clean_name.startswith(p):
                            clean_name = clean_name[len(p):]
                            break

                if clean_name in exclude_property_names_after_remove_prefix or clean_name in res:
                    continue

                res.append(clean_name)

        return res

    def simple_dict(
            self,
            *,
            need_include_columns: bool = True,
            need_include_sd_properties: bool = True,
            include_columns: Collection[str] | None = None,
            exclude_columns: Collection[str] | None = None,
            include_sd_properties: Collection[str] | None = None,
            exclude_sd_properties: Collection[str] | None = None,
            include_columns_and_sd_properties: Collection[str] | None = None,
            include_sd_property_prefixes: list[str] | None = None,
            remove_sd_property_prefixes: list[str] | None = None,
            kwargs: dict[str, Any] | None = None
    ) -> dict[str, Any]:
        """
        Преобразует ORM-объект в простой dict, включая обычные колонки и sd-свойства.

        :param need_include_columns: Включать ли обычные колонки модели.
        :param need_include_sd_properties: Включать ли sd-свойства (property с заданными префиксами).
        :param include_columns: Список колонок для включения.
        :param exclude_columns: Список колонок для исключения.
        :param include_sd_properties: Список sd-свойств для включения (по полным именам).
        :param exclude_sd_properties: Список sd-свойств для исключения (по полным именам).
        :param include_columns_and_sd_properties: Универсальный фильтр — если задан, то включаются только эти имена.
        :param include_sd_property_prefixes: Список префиксов, которые считаются sd-свойствами. По умолчанию ["sdp_"].
        :param remove_sd_property_prefixes: Префиксы, которые нужно удалить из имени при добавлении в результат.
        :param kwargs: Дополнительные данные, которые будут добавлены в итоговый словарь.
        """
        exclude_columns = set(exclude_columns or [])
        exclude_sd_properties = set(exclude_sd_properties or [])
        if include_sd_property_prefixes is None:
            include_sd_property_prefixes = include_sd_property_prefixes or ["sdp_"]
        if remove_sd_property_prefixes is None:
            remove_sd_property_prefixes = remove_sd_property_prefixes or ["sdp_"]

        res = {}

        # === 1. Обычные колонки ===
        if need_include_columns:
            for c in inspect(self).mapper.column_attrs:
                if include_columns_and_sd_properties is not None and c.key not in include_columns_and_sd_properties:
                    continue
                if include_columns is not None and c.key not in include_columns:
                    continue
                if c.key in exclude_columns:
                    continue
                value = getattr(self, c.key)
                res[c.key] = value

        # === 2. SD-свойства ===
        if need_include_sd_properties:
            for attr_name in dir(self):
                # Проверяем, начинается ли имя с любого префикса
                if not any(attr_name.startswith(p) for p in include_sd_property_prefixes):
                    continue
                if not isinstance(getattr(type(self), attr_name, None), property):
                    continue

                sd_property_name = attr_name  # пока полное имя (с префиксом)
                # сначала убираем префикс (первый совпавший)
                for p in remove_sd_property_prefixes:
                    if sd_property_name.startswith(p):
                        sd_property_name = sd_property_name[len(p):]
                        break

                # Фильтрация — ДО удаления префикса
                if (
                        include_columns_and_sd_properties is not None
                        and sd_property_name not in include_columns_and_sd_properties
                ):
                    continue
                if include_sd_properties is not None and sd_property_name not in include_sd_properties:
                    continue
                if sd_property_name in exclude_sd_properties:
                    continue

                res[sd_property_name] = getattr(self, attr_name)

        # === 3. Добавляем дополнительные данные ===
        if kwargs is not None:
            res.update(kwargs)

        return res


class SQLAlchemyDb:
    def __init__(
            self,
            *,
            sync_db_url: str | None = "postgresql://arpakitlib:arpakitlib@localhost:50517/arpakitlib",
            async_db_url: str | None = "postgresql+asyncpg://arpakitlib:arpakitlib@localhost:50517/arpakitlib",
            db_echo: bool = False,
            base_dbm: type[BaseDBM] | None = None,
            db_models: list[Any] | None = None,
    ):
        self._logger = logging.getLogger(self.__class__.__name__)

        self.db_url = sync_db_url
        if self.db_url is not None:
            self.engine = create_engine(
                url=sync_db_url,
                echo=db_echo,
                pool_size=20,
                max_overflow=30,
                poolclass=QueuePool,
                pool_timeout=timedelta(seconds=30).total_seconds(),
            )
        self.sessionmaker = sessionmaker(bind=self.engine)
        self.func_new_session_counter = 0

        self.async_db_url = async_db_url
        if self.async_db_url is not None:
            self.async_engine = create_async_engine(
                url=async_db_url,
                echo=db_echo,
                pool_size=20,
                max_overflow=30,
                poolclass=AsyncAdaptedQueuePool,
                pool_timeout=timedelta(seconds=30).total_seconds()
            )
        self.async_sessionmaker = async_sessionmaker(bind=self.async_engine)
        self.func_new_async_session_counter = 0

        self.base_dbm = base_dbm
        self.db_models = db_models

    def is_table_exists(self, table_name: str) -> bool:
        with self.engine.connect() as connection:
            inspector = inspect(connection)
            return table_name in inspector.get_table_names()

    def drop_celery_tables(self):
        with self.engine.connect() as connection:
            connection.execute(text("DROP TABLE IF EXISTS celery_tasksetmeta CASCADE;"))
            connection.execute(text("DROP TABLE IF EXISTS celery_taskmeta CASCADE;"))
            connection.commit()
        self._logger.info("celery tables were dropped")

    def remove_celery_tables_data(self):
        if not self.is_table_exists("celery_tasksetmeta"):
            self._logger.info("table celery_tasksetmeta not exists")
            return
        with self.engine.connect() as connection:
            connection.execute(text("DELETE FROM celery_tasksetmeta;"))
            connection.execute(text("DELETE FROM celery_taskmeta;"))
            connection.commit()
        self._logger.info("celery tables data were removed")

    def drop_alembic_tables(self):
        with self.engine.connect() as connection:
            connection.execute(text("DROP TABLE IF EXISTS alembic_version CASCADE;"))
            connection.execute(text("DROP TABLE IF EXISTS alembic_version CASCADE;"))
            connection.commit()
        self._logger.info("alembic_version tables were dropped")

    def remove_alembic_tables_data(self):
        if not self.is_table_exists("alembic_version"):
            self._logger.info("table alembic_version not exists")
            return
        with self.engine.connect() as connection:
            connection.execute(text("DELETE FROM alembic_version;"))
            connection.commit()
        self._logger.info("alembic tables data were removed")

    def ensure_check_constraints(self):
        from arpakitlib.ar_sqlalchemy_ensure_check_constraints_util import ensure_sqlalchemy_check_constraints
        ensure_sqlalchemy_check_constraints(base_=self.base_dbm, engine=self.engine)

    def drop_check_constraints(self):
        from arpakitlib.ar_sqlalchemy_drop_check_constraints_util import drop_sqlalchemy_check_constraints
        drop_sqlalchemy_check_constraints(base_=self.base_dbm, engine=self.engine)

    def drop_and_ensure_check_constraints(self):
        self.drop_check_constraints()
        self.ensure_check_constraints()

    def init(self, ensure_check_constraints: bool = True):
        self.base_dbm.metadata.create_all(bind=self.engine, checkfirst=True)
        if ensure_check_constraints:
            self.ensure_check_constraints()
        self._logger.info("inited")

    def drop(self):
        self.base_dbm.metadata.drop_all(bind=self.engine, checkfirst=True)
        self._logger.info("dropped")

    def reinit(self, ensure_check_constraints: bool = True):
        self.base_dbm.metadata.drop_all(bind=self.engine, checkfirst=True)
        self.base_dbm.metadata.create_all(bind=self.engine, checkfirst=True)
        if ensure_check_constraints:
            self.ensure_check_constraints()
        self._logger.info("reinited")

    def reinit_all(self):
        self.reinit()
        self.remove_alembic_tables_data()
        self.remove_celery_tables_data()
        self._logger.info("all reinited")

    def remove_rows_from_tables(self, exclude_tables: list[str] | None = None):
        if exclude_tables is None:
            exclude_tables = []
        with self.new_session() as session:
            for table_name, table in BaseDBM.metadata.tables.items():
                if table_name not in exclude_tables:
                    session.execute(sqlalchemy.delete(table))
            session.commit()
        removed_tables = [t for t in BaseDBM.metadata.tables.keys() if t not in exclude_tables]
        self._logger.info(f"rows from tables ({removed_tables}) were removed")

    async def async_remove_rows_from_tables(self, exclude_tables: list[str] | None = None):
        if exclude_tables is None:
            exclude_tables = []
        async with self.new_async_session() as async_session:
            for table_name, table in BaseDBM.metadata.tables.items():
                if table_name not in exclude_tables:
                    await async_session.execute(sqlalchemy.delete(table))
            await async_session.commit()
        removed_tables = [t for t in BaseDBM.metadata.tables.keys() if t not in exclude_tables]
        self._logger.info(f"rows from tables ({removed_tables}) were removed")

    def check_conn(self):
        self.engine.connect()
        self._logger.info("db conn is good")

    def new_session(self, **kwargs) -> Session:
        self.func_new_session_counter += 1
        return self.sessionmaker(**kwargs)

    def new_async_session(self, **kwargs) -> AsyncSession:
        self.func_new_async_session_counter += 1
        return self.async_sessionmaker(**kwargs)

    def is_conn_good(self) -> bool:
        try:
            self.check_conn()
        except Exception as e:
            return False
        return True

    def generate_unique_id(self, *, class_dbm: Any):
        with self.new_session() as session:
            res: int = session.query(func.max(class_dbm.id)).scalar()
            while session.query(class_dbm).filter(class_dbm.id == res).first() is not None:
                res += 1
        return res

    def generate_unique_long_id(self, *, class_dbm: Any):
        with self.new_session() as session:
            res: str = str(uuid4())
            while session.query(class_dbm).filter(class_dbm.long_id == res).first() is not None:
                res = str(uuid4())
        return res

    def generate_creation_dt(self) -> datetime:
        return now_utc_dt()

    async def async_get_table_name_to_amount(self) -> dict[str, int]:
        res = {}

        async with self.new_async_session() as async_session:
            for table_name, table in BaseDBM.metadata.tables.items():
                res[table_name] = await async_session.scalar(
                    sqlalchemy.select(
                        sqlalchemy.func.count(1)
                    ).select_from(table)
                )

        return res

    def get_table_name_to_amount(self) -> dict[str, int]:
        res = {}

        with self.new_session() as session:
            for table_name, table in BaseDBM.metadata.tables.items():
                res[table_name] = session.scalar(
                    sqlalchemy.select(
                        sqlalchemy.func.count(1)
                    ).select_from(table)
                )

        return res


def get_string_info_from_declarative_base(class_: type[DeclarativeBase]):
    res = f"Db Models: {len(class_.__subclasses__())}"
    for i, cls in enumerate(class_.__subclasses__()):
        res += f"\n{i + 1}. {cls.__name__}"
    return res


def get_string_info_from_declarative_base_2(class_: type[DeclarativeBase]) -> str:
    lines: list[str] = [f"DBModels: {len(class_.__subclasses__())}"]
    for i, cls in enumerate(class_.__subclasses__()):
        if hasattr(cls, "__tablename__"):
            mapper = inspect(cls)
            lines.append(f"{i + 1}. DBModel: {cls.__tablename__} (columns={len(mapper.columns)})")
            for column in mapper.columns:
                lines.append(f"   - {column.name} ({column.type})")
    return "\n".join(lines)


def __example():
    pass


async def __async_example():
    pass


if __name__ == '__main__':
    __example()
    asyncio.run(__async_example())



from __future__ import annotations

import asyncio
import logging
import uuid
from datetime import timedelta
from typing import Any

import aiohttp
import requests

from arpakitlib.ar_base_http_api_client_util import BaseHTTPAPIClient
from arpakitlib.ar_dict_util import combine_dicts
from arpakitlib.ar_enumeration_util import Enumeration
from arpakitlib.ar_http_request_util import sync_make_http_request, async_make_http_request
from arpakitlib.ar_type_util import raise_for_type

"""
https://yookassa.ru/developers/api
"""


class EasyYookassaAPIClient(BaseHTTPAPIClient):
    class PaymentStatuses(Enumeration):
        pending = "pending"
        waiting_for_capture = "waiting_for_capture"
        succeeded = "succeeded"
        canceled = "canceled"

    def __init__(self, *, secret_key: str, shop_id: int):
        super().__init__()
        self.secret_key = secret_key
        self.shop_id = shop_id
        self.headers = {"Content-Type": "application/json"}
        self._logger = logging.getLogger(f"{self.__class__.__name__}-{shop_id}")

    def _sync_make_http_request(
            self,
            *,
            method: str,
            url: str,
            headers: dict[str, Any] | None = None,
            not_raise_for_statuses_: list[int] | set[int] | None = None,
            **kwargs
    ) -> requests.Response:
        return sync_make_http_request(
            method=method,
            url=url,
            headers=combine_dicts(self.headers, (headers if headers is not None else {})),
            max_tries_=1,
            raise_for_status_=True,
            timeout_=timedelta(seconds=5),
            not_raise_for_statuses_=not_raise_for_statuses_,
            auth=(self.shop_id, self.secret_key),
            enable_logging_=False,
            **kwargs
        )

    async def _async_make_http_request(
            self,
            *,
            method: str = "GET",
            url: str,
            headers: dict[str, Any] | None = None,
            not_raise_for_statuses_: list[int] | set[int] | None = None,
            **kwargs
    ) -> aiohttp.ClientResponse:
        return await async_make_http_request(
            method=method,
            url=url,
            headers=combine_dicts(self.headers, (headers if headers is not None else {})),
            max_tries_=1,
            raise_for_status_=True,
            not_raise_for_statuses_=not_raise_for_statuses_,
            timeout_=timedelta(seconds=5),
            auth=aiohttp.BasicAuth(login=str(self.shop_id), password=self.secret_key),
            enable_logging_=False,
            **kwargs
        )

    def sync_create_payment(self, *, json_body: dict[str, Any]) -> dict[str, Any]:

        """
        json_body example
        json_body = {
            "amount": {
                "value": "2.0",
                "currency": "RUB"
            },
            "description": "description",
            "confirmation": {
                "type": "redirect",
                "return_url": f"https://t.me/{get_tg_bot_username()}",
                "locale": "ru_RU"
            },
            "capture": True,
            "metadata": {},
            "merchant_customer_id": ""
        }
        """

        response = self._sync_make_http_request(
            method="POST",
            url="https://api.yookassa.ru/v3/payments",
            headers={"Idempotence-Key": str(uuid.uuid4())},
            json=json_body,
        )
        json_data = response.json()
        response.raise_for_status()
        return json_data

    def sync_get_payment(self, *, payment_id: str) -> dict[str, Any] | None:
        raise_for_type(payment_id, str)
        response = self._sync_make_http_request(
            method="GET",
            url=f"https://api.yookassa.ru/v3/payments/{payment_id}",
            not_raise_for_statuses_=[404]
        )
        json_data = response.json()
        if response.status_code == 404:
            return None
        response.raise_for_status()
        return json_data

    async def async_create_payment(self, *, json_body: dict[str, Any]) -> dict[str, Any]:

        """
        json_body example
        json_body = {
            "amount": {
                "value": "2.0",
                "currency": "RUB"
            },
            "description": "description",
            "confirmation": {
                "type": "redirect",
                "return_url": f"https://t.me/{get_tg_bot_username()}",
                "locale": "ru_RU"
            },
            "capture": True,
            "metadata": {},
            "merchant_customer_id": ""
        }
        """

        response = await self._async_make_http_request(
            method="POST",
            url="https://api.yookassa.ru/v3/payments",
            headers={"Idempotence-Key": str(uuid.uuid4())},
            json=json_body,
        )
        json_data = await response.json()
        response.raise_for_status()
        return json_data

    async def async_get_payment(self, *, payment_id: str) -> dict[str, Any] | None:
        raise_for_type(payment_id, str)
        response = await self._async_make_http_request(
            method="GET",
            url=f"https://api.yookassa.ru/v3/payments/{payment_id}",
            not_raise_for_statuses_=[404]
        )
        json_data = await response.json()
        if response.status == 404:
            return None
        response.raise_for_status()
        return json_data


YookassaAPIClient = EasyYookassaAPIClient


def __example():
    pass


async def __async_example():
    pass


if __name__ == '__main__':
    __example()
    asyncio.run(__async_example())



# arpakit

from datetime import datetime, date
from zoneinfo import ZoneInfo

import pytz

from arpakitlib.ar_type_util import raise_for_type


def convert_dt_tz(dt: datetime, tz_info):
    return dt.astimezone(tz_info)


def convert_dt_tz_to_utc(dt: datetime):
    return convert_dt_tz(dt=dt, tz_info=pytz.UTC)


def now_utc_dt() -> datetime:
    return datetime.now(tz=pytz.UTC)


def now_dt(tz=pytz.UTC) -> datetime:
    return datetime.now(tz=tz)


def birth_date_to_age(*, birth_date: date, raise_if_age_negative: bool = False) -> int:
    raise_for_type(birth_date, date)
    now_utc_dt_date = now_utc_dt().date()
    res = now_utc_dt_date.year - birth_date.year
    if (now_utc_dt_date.month, now_utc_dt_date.day) < (birth_date.month, birth_date.day):
        res -= 1
    if raise_if_age_negative and res < 0:
        raise ValueError("raise_if_negative and res < 0")
    return res


def datetime_as_msk_str(datetime_: datetime | None) -> str | None:
    if datetime_ is None:
        return None
    return datetime_.astimezone(ZoneInfo("Europe/Moscow")).strftime("%Y-%m-%d %H:%M:%S %p %Z %z")


def datetime_as_ufa_str(datetime_: datetime | None) -> str | None:
    if datetime_ is None:
        return None
    return datetime_.astimezone(ZoneInfo("Asia/Yekaterinburg")).strftime("%Y-%m-%d %H:%M:%S %p %Z %z")


def __example():
    pass


if __name__ == '__main__':
    __example()



import os


def raise_if_path_not_exists(path: str):
    if not os.path.exists(path):
        raise Exception(f"path {path} not exists")


def clear_text_file(filepath: str):
    if not os.path.exists(filepath):
        return
    with open(filepath, mode="w") as f:
        f.write("")


def create_text_file(filepath: str, text: str = ""):
    if os.path.exists(filepath):
        return
    with open(filepath, mode="w") as f:
        f.write(text)



# arpakit
import os
import shlex

from pydantic import BaseModel


class BadCommandFormat(Exception):
    pass


class ParsedCommand(BaseModel):
    command: str
    full_command: str
    key_to_value: dict[str, str | None] = {}
    values_without_key: list[str] = []

    @property
    def keys(self) -> list[str]:
        return [k for k, v in self.key_to_value.items() if v is not None]

    @property
    def flags(self) -> list[str]:
        return [k for k, v in self.key_to_value.items() if v is None]

    @property
    def values(self) -> list[str]:
        return [self.key_to_value[k] for k in self.keys]

    def get_value_by_key(self, key: str) -> str | None:
        return self.key_to_value.get(key)

    def get_value_by_keys(self, keys: list[str]) -> str | None:
        for key in keys:
            if self.key_exists(key=key):
                return self.get_value_by_key(key=key)
        return None

    def key_exists(self, key: str) -> bool:
        return key in self.key_to_value.keys()

    def keys_exists(self, keys: list[str]) -> bool:
        for key in keys:
            if key in self.keys:
                return True
        return False

    def has_flag(self, flag: str) -> bool:
        return flag in self.flags

    def get_value_by_index(self, index: int) -> str | None:
        if index >= len(self.values_without_key):
            return None
        return self.values_without_key[index]


def parse_command(text: str) -> ParsedCommand:
    text = text.removeprefix("/")
    text = " ".join([text_.strip() for text_ in text.split(" ") if text_.strip()]).strip()

    parts = shlex.split(text)
    if not parts:
        raise BadCommandFormat("not parts")
    if len(parts[0]) == 1:
        raise BadCommandFormat("len(parts[0]) == 1")

    res = ParsedCommand(full_command=parts[0], command=os.path.basename(parts[0]).removeprefix("/"))

    last_key: str | None = None
    for part in parts[1:]:
        part = part.strip()

        if not part:
            raise BadCommandFormat("not part")
        if part == "-" or part == "--":
            raise BadCommandFormat("part == '-' or part == '--'")

        if part.startswith("-") or part.startswith("--"):  # if it is key
            if part.startswith("-"):
                part = part[1:]
            if part.startswith("-"):
                part = part[1:]
            if part.startswith("-"):
                raise BadCommandFormat("a lots of -")

            if part in res.key_to_value:
                raise BadCommandFormat(f"{part} in {res.key_to_value}")

            res.key_to_value[part] = None
            last_key = part

            continue

        if last_key is not None:  # if it is value
            res.key_to_value[last_key] = part
            last_key = None
            continue

        res.values_without_key.append(part)  # if it is values_without_key

    return res


def __example():
    pass


if __name__ == '__main__':
    __example()



# arpakit

from shlex import quote

from arpakitlib.ar_run_cmd_util import run_cmd
from arpakitlib.ar_type_util import raise_for_type


def make_postgresql_db_dump(
        *,
        user: str,
        host: str = "127.0.0.1",
        db_name: str,
        port: int = 5432,
        out_filepath: str = "db_dump.sql",
        password: str | None = None,
) -> str:
    raise_for_type(user, str)
    raise_for_type(host, str)
    raise_for_type(db_name, str)
    raise_for_type(port, int)
    raise_for_type(out_filepath, str)

    env_prefix = ""
    if password:
        env_prefix = f"PGPASSWORD={quote(password)} "

    command = (
        f"{env_prefix}"
        f"pg_dump "
        f"-U {quote(user)} "
        f"-h {quote(host)} "
        f"-p {port} "
        f"-f {quote(out_filepath)} "
        f"{quote(db_name)}"
    )

    res = run_cmd(command)
    res.raise_for_bad_return_code()

    return out_filepath


def __example():
    pass


if __name__ == '__main__':
    __example()



import random
import secrets
import string
import uuid


def generate_difficult_password(*, difficult: int = 1):
    # динамический диапазон длины
    base = 32 + difficult * 16
    variance = 64 + difficult * 32
    target_len = random.randint(base, base + variance)

    # ТОЛЬКО английские буквы и цифры
    alphabet = string.ascii_letters + string.digits

    password = []

    # UUID-шум (hex = 0-9 + a-f)
    for _ in range(difficult * 2):
        password.append(uuid.uuid4().hex)

    # основная часть
    for _ in range(target_len):
        password.append(secrets.choice(alphabet))

    secrets.SystemRandom().shuffle(password)

    return "".join(password)


def __example():
    print(generate_difficult_password(difficult=2))


if __name__ == '__main__':
    __example()



# arpakit

import base64
from typing import Optional


def convert_base64_string_to_bytes(base64_string: str, raise_for_error: bool = False) -> Optional[bytes]:
    try:
        return base64.b64decode(base64_string)
    except Exception as e:
        if raise_for_error:
            raise e
        return None


def convert_base64_string_to_string(base64_string: str, raise_for_error: bool = False) -> Optional[str]:
    return convert_base64_string_to_bytes(
        base64_string=base64_string,
        raise_for_error=raise_for_error
    ).decode()


def convert_bytes_to_base64_string(bytes_: bytes, raise_for_error: bool = False) -> Optional[str]:
    try:
        return base64.b64encode(bytes_).decode()
    except Exception as e:
        if raise_for_error:
            raise e
        return None


def convert_file_to_base64_string(*, filepath: str, raise_for_error: bool = False) -> Optional[str]:
    with open(filepath, "rb") as f:
        return convert_bytes_to_base64_string(bytes_=f.read(), raise_for_error=raise_for_error)


def __example():
    pass


if __name__ == '__main__':
    __example()



# arpakit

import logging
import os
import shutil
import uuid
from typing import Optional, Union, Iterator

from arpakitlib.ar_datetime_util import now_utc_dt
from arpakitlib.ar_str_util import none_if_blank


class FileStorageInDir:

    def __init__(self, dirpath: str):
        if not isinstance(dirpath, str):
            raise TypeError("not isinstance(dirpath, str)")
        self.dirpath = os.path.abspath(dirpath)
        self._logger = logging.getLogger(self.__class__.__name__)

    def __str__(self) -> str:
        return self.dirpath

    def __repr__(self) -> str:
        return self.dirpath

    def __len__(self) -> int:
        return len(os.listdir(self.dirpath))

    def __iter__(self) -> Iterator[str]:
        self.init()
        for path in self.get_paths():
            yield path

    def init(self):
        if not os.path.exists(self.dirpath):
            os.makedirs(self.dirpath, exist_ok=True)

    def reinit(self):
        if os.path.exists(self.dirpath):
            shutil.rmtree(self.dirpath)
        self.init()

    def get_paths(self) -> list[str]:
        return [os.path.join(self.dirpath, path) for path in os.listdir(self.dirpath)]

    def generate_filepath(
            self,
            *,
            filename: Optional[str] = None,
            file_extension: Optional[str] = None,
            content_to_write: Optional[Union[str, bytes]] = None,
            create: bool = False,
            raise_if_exists: bool = True,
            add_datetime_in_filename: bool = False
    ) -> str:
        self.init()

        filename = none_if_blank(filename)

        if filename is not None:
            if add_datetime_in_filename:
                filename = f"{now_utc_dt().isoformat()}--{filename}"
            if file_extension is not None:
                filename += f".{file_extension}"
        else:
            filename = str(uuid.uuid4())
            if add_datetime_in_filename:
                filename = f"{now_utc_dt().isoformat()}--{filename}"
            if file_extension is not None:
                filename += f".{file_extension}"
            while filename in os.listdir(self.dirpath):
                filename = str(uuid.uuid4())
                if file_extension is not None:
                    filename += f".{file_extension}"

        if raise_if_exists and filename in os.listdir(self.dirpath):
            raise ValueError(f"file ({filename}) already exists")

        filepath = os.path.join(self.dirpath, filename)

        if create is True:
            content_to_write = ""

        if isinstance(content_to_write, str):
            content_to_write = content_to_write.encode()
        if isinstance(content_to_write, bytes):
            with open(filepath, mode="wb") as f:
                f.write(content_to_write)

        return filepath

    def generate_dirpath(
            self, *, dirname: Optional[str] = None, create: bool = True, raise_if_exists: bool = True
    ) -> str:
        self.init()

        if dirname is None:
            dirname = str(uuid.uuid4())
            while dirname in os.listdir(self.dirpath):
                dirname = str(uuid.uuid4())

        if raise_if_exists and dirname in os.listdir(self.dirpath):
            raise ValueError(f"dir ({dirname}) already exists")

        dirpath = os.path.join(self.dirpath, dirname)

        if create:
            os.mkdir(dirpath)

        return dirpath

    def rm_by_filepath(self, filepath: str):
        if os.path.exists(filepath):
            os.remove(filepath)

    def rm_by_filename(self, filename: str):
        filepath = os.path.join(self.dirpath, filename)
        if os.path.exists(filepath):
            os.remove(filepath)

    def rm_by_dirpath(self, dirpath: str):
        if os.path.exists(dirpath):
            shutil.rmtree(dirpath)

    def rm_by_dirname(self, dirname: str):
        dirpath = os.path.join(self.dirpath, dirname)
        if os.path.exists(dirpath):
            shutil.rmtree(dirpath)


def __example():
    pass


if __name__ == '__main__':
    __example()
